from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_properties import SDL_PropertiesID
from .SDL_keycode import SDL_Keycode, SDL_Keymod
from .SDL_scancode import SDL_Scancode
from .SDL_video import SDL_Window
from .SDL_rect import SDL_Rect

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_KeyboardID = ctypes.c_uint32

SDL_FUNC("SDL_HasKeyboard", ctypes.c_bool)
SDL_FUNC("SDL_GetKeyboards", ctypes.POINTER(SDL_KeyboardID), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetKeyboardNameForID", ctypes.c_char_p, SDL_KeyboardID)
SDL_FUNC("SDL_GetKeyboardFocus", ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_GetKeyboardState", ctypes.POINTER(ctypes.c_bool), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_ResetKeyboard", None)
SDL_FUNC("SDL_GetModState", SDL_Keymod)
SDL_FUNC("SDL_SetModState", None, SDL_Keymod)
SDL_FUNC("SDL_GetKeyFromScancode", SDL_Keycode, SDL_Scancode, SDL_Keymod, ctypes.c_bool)
SDL_FUNC("SDL_GetScancodeFromKey", SDL_Scancode, SDL_Keycode, ctypes.POINTER(SDL_Keymod))
SDL_FUNC("SDL_SetScancodeName", ctypes.c_bool, SDL_Scancode, ctypes.c_char_p)
SDL_FUNC("SDL_GetScancodeName", ctypes.c_char_p, SDL_Scancode)
SDL_FUNC("SDL_GetScancodeFromName", SDL_Scancode, ctypes.c_char_p)
SDL_FUNC("SDL_GetKeyName", ctypes.c_char_p, SDL_Keycode)
SDL_FUNC("SDL_GetKeyFromName", SDL_Keycode, ctypes.c_char_p)

SDL_FUNC("SDL_StartTextInput", ctypes.c_bool, ctypes.POINTER(SDL_Window))

SDL_TextInputType = ctypes.c_int

SDL_TEXTINPUT_TYPE_TEXT = 0
SDL_TEXTINPUT_TYPE_TEXT_NAME = 1
SDL_TEXTINPUT_TYPE_TEXT_EMAIL = 2
SDL_TEXTINPUT_TYPE_TEXT_USERNAME = 3
SDL_TEXTINPUT_TYPE_TEXT_PASSWORD_HIDDEN = 4
SDL_TEXTINPUT_TYPE_TEXT_PASSWORD_VISIBLE = 5
SDL_TEXTINPUT_TYPE_NUMBER = 6
SDL_TEXTINPUT_TYPE_NUMBER_PASSWORD_HIDDEN = 7
SDL_TEXTINPUT_TYPE_NUMBER_PASSWORD_VISIBLE = 8

SDL_Capitalization = ctypes.c_int

SDL_CAPITALIZE_NONE = 0
SDL_CAPITALIZE_SENTENCES = 1
SDL_CAPITALIZE_WORDS = 2
SDL_CAPITALIZE_LETTERS = 3

SDL_FUNC("SDL_StartTextInputWithProperties", ctypes.c_bool, ctypes.POINTER(SDL_Window), SDL_PropertiesID)

SDL_PROP_TEXTINPUT_TYPE_NUMBER = "SDL.textinput.type".encode()
SDL_PROP_TEXTINPUT_CAPITALIZATION_NUMBER = "SDL.textinput.capitalization".encode()
SDL_PROP_TEXTINPUT_AUTOCORRECT_BOOLEAN = "SDL.textinput.autocorrect".encode()
SDL_PROP_TEXTINPUT_MULTILINE_BOOLEAN = "SDL.textinput.multiline".encode()
SDL_PROP_TEXTINPUT_ANDROID_INPUTTYPE_NUMBER = "SDL.textinput.android.inputtype".encode()

SDL_FUNC("SDL_TextInputActive", ctypes.c_bool, ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_StopTextInput", ctypes.c_bool, ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_ClearComposition", ctypes.c_bool, ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_SetTextInputArea", ctypes.c_bool, ctypes.POINTER(SDL_Window), ctypes.POINTER(SDL_Rect), ctypes.c_int)
SDL_FUNC("SDL_GetTextInputArea", ctypes.c_bool, ctypes.POINTER(SDL_Window), ctypes.POINTER(SDL_Rect), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_HasScreenKeyboardSupport", ctypes.c_bool)
SDL_FUNC("SDL_ScreenKeyboardShown", ctypes.c_bool, ctypes.POINTER(SDL_Window))


from .__init__ import sdl3, TEST_RegisterFunction

@TEST_RegisterFunction(["Darwin", "Windows"])
def TEST_SDL_CreateWindow() -> None:
    assert sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO), sdl3.SDL_GetError().decode()
    assert (window := sdl3.SDL_CreateWindow("Test".encode(), 1600, 900, sdl3.SDL_WINDOW_RESIZABLE)), sdl3.SDL_GetError().decode()
    assert (error := sdl3.SDL_GetError()) == "".encode(), error.decode()
    sdl3.SDL_DestroyWindow(window)
    sdl3.SDL_Quit()


"""A pure Python wrapper for SDL3."""

__version__ = "0.9.4b4"

import sys, os, requests, ctypes, ctypes.util, platform, keyword, inspect, types, \
    asyncio, aiohttp, zipfile, typing, array, atexit, packaging.version, json, re

SDL_BINARY, SDL_IMAGE_BINARY, SDL_MIXER_BINARY, SDL_TTF_BINARY, SDL_RTF_BINARY, SDL_NET_BINARY = \
    "SDL3", "SDL3_image", "SDL3_mixer", "SDL3_ttf", "SDL3_rtf", "SDL3_net"

SDL_BINARY_VAR_MAP: typing.Dict[str, str] = {}

for i in locals().copy():
    if i.startswith("SDL_") and i.endswith("_BINARY"):
        SDL_BINARY_VAR_MAP[i] = locals()[i]

SDL_BINARY_VAR_MAP_INV: typing.Dict[str, str] = {value: key for key, value in SDL_BINARY_VAR_MAP.items()}
SDL_REPOSITORIES: typing.List[str] = [key.replace("3", "") for key, _ in SDL_BINARY_VAR_MAP_INV.items()]

def SDL_FORMAT_ARCH(arch: str) -> str:
    """Format the architecture string."""
    if arch.lower() in ["x64", "x86_64", "amd64"]: return "AMD64"
    if arch.lower() in ["aarch64", "arm64"]: return "ARM64"
    assert False, "Unknown architecture."

SDL_SYSTEM, SDL_ARCH = platform.system(), SDL_FORMAT_ARCH(platform.machine())
SDL_BINARY_PATTERNS: typing.Dict[str, typing.List[str]] = \
    {"Windows": ["{}.dll"], "Darwin": ["lib{}.dylib", "{0}.framework/{0}", "{0}.framework/Versions/A/{0}"], "Linux": ["lib{}.so"]}

__frozen__: bool = getattr(sys, "frozen", False)
__doc_file__: str = os.path.join(os.path.dirname(__file__), "__doc__.py")
__doc_generator__: int = int(os.environ.get("SDL_DOC_GENERATOR", str(int(not __frozen__)))) > 0

__initialized__: bool = __name__.split(".")[0] in sys.modules if "." in __name__ else False
__module__: types.ModuleType = sys.modules[__name__.split(".")[0] if "." in __name__ else __name__]

def SDL_FIND_BINARIES(libraries: typing.List[str]) -> typing.List[str]:
    """Find binaries of system libraries."""
    libraries = libraries + [f"{library}d" for library in libraries]
    binaries = [f"./{file}" if SDL_SYSTEM in ["Windows"] and not ("/" in file or "\\" in file) else file for library in libraries if (file := ctypes.util.find_library(library))]

    if SDL_SYSTEM in ["Darwin"] and SDL_ARCH in ["ARM64"] and os.path.exists(path := "/opt/Homebrew/lib"):
        binaries += [file for library in libraries for pattern in SDL_BINARY_PATTERNS if os.path.exists(file := os.path.join(path, pattern.format(library)))]

    return binaries

def SDL_DOWNLOAD_BINARIES(path: str, system: str = SDL_SYSTEM, arch: str = SDL_ARCH) -> None:
    """Download specific SDL3 binaries to the given path."""
    assert system.capitalize() in ["Darwin", "Linux", "Windows"], "Unknown system."
    assert arch.upper() in ["AMD64", "ARM64"], "Unknown architecture."
    if not os.path.exists(path): os.makedirs(path)

    headers = {"Accept": "application/vnd.github+json"}

    if "SDL_GITHUB_TOKEN" in os.environ:
        headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

    try:
        for release in requests.get("https://api.github.com/repos/Aermoss/PySDL3-Build/releases", headers = headers).json():
            if release["draft"] or release["prerelease"]:
                continue

            for asset in release["assets"]:
                if asset["name"] != f"{system.capitalize()}-{arch.upper()}-{release['tag_name']}.zip":
                    continue

                with requests.get(asset["browser_download_url"], headers = headers, stream = True) as response:
                    assert response.status_code == 200, f"failed to get binaries from github, status: {response.status_code}."
                    size, current = int(response.headers.get("content-length", 0)), 0

                    with open(os.path.join(path, asset["name"]), "wb") as file:
                        for chunk in response.iter_content(chunk_size = 8192):
                            if chunk: file.write(chunk); current += len(chunk)
                            print("\33[32m", f"info: downloading '{asset['browser_download_url']}'... {current / size * 100:.1f}% ({current / (1024 ** 2):.2f}/{size / (1024 ** 2):.2f} MB).", "\33[0m", sep = "", end = "\r", flush = True)

                        print(flush = True)

                    data = {
                        "arch": arch, "system": system, "target": f"v{__version__}",
                        "version": release["tag_name"], "url": asset["browser_download_url"],
                        "created": asset["created_at"], "updated": asset["updated_at"],
                        "uploader": asset["uploader"]["login"], "files": [],
                        "repair": True, "find": True
                    }

                    with zipfile.ZipFile(os.path.join(path, asset["name"]), "r") as ref:
                        for name in ref.namelist():
                            data["files"].append(f"./{name}")

                            if os.path.exists(os.path.join(path, name)):
                                os.remove(os.path.join(path, name))

                        ref.extractall(path)

                    with open(os.path.join(path, "metadata.json"), "w") as file:
                        json.dump(data, file, indent = 4)

                    os.remove(os.path.join(path, asset["name"]))
                    return
                
    except requests.RequestException as exc:
        print("\33[31m", f"error: failed to download binaries: {str(exc).lower()}", "\33[0m", sep = "", flush = True)

if not __initialized__:
    if int(os.environ.get("SDL_CHECK_VERSION", str(int(not __frozen__)))) > 0:
        try:
            version = requests.get(f"https://pypi.org/pypi/PySDL3/json", timeout = 0.5).json()["info"]["version"]

            if packaging.version.parse(__version__) < packaging.version.parse(version):
                print("\33[35m", f"warning: you are using an older version of pysdl3 (current: {__version__}, lastest: {version}).", "\33[0m", sep = "", flush = True)

        except requests.RequestException:
            ...

    functions, binaryMap, currentBinary, missing = {}, {}, None, True
    binaryData = {"system": SDL_SYSTEM, "arch": SDL_ARCH, "files": []}
    binaryPath = os.environ.get("SDL_BINARY_PATH", os.path.join(os.path.dirname(__file__), "bin"))
    absPath = lambda path: path if os.path.isabs(path) else os.path.abspath(os.path.join(binaryPath, path))
    if not os.path.exists(binaryPath): os.makedirs(binaryPath)

    if "metadata.json" in (files := os.listdir(binaryPath)):
        with open(os.path.join(binaryPath, "metadata.json"), "r") as file:
            binaryData = json.load(file)
            binaryData["files"] = [absPath(i) for i in binaryData["files"]]

        if packaging.version.parse(__version__) > packaging.version.parse(binaryData.get("target", __version__)):
            print("\33[35m", f"warning: incompatible target version detected: '{binaryData['target']}', current: 'v{__version__}'.", "\33[0m", sep = "", flush = True)

        elif binaryData["system"].lower() != SDL_SYSTEM.lower() or binaryData["arch"].lower() != SDL_ARCH.lower():
            print("\33[35m", f"warning: incompatible binary architecture and/or system detected: '{binaryData['system']} ({binaryData['arch']})'.", "\33[0m", sep = "", flush = True)
            binaryData["repair"] = True

            for i in binaryData["files"]:
                if os.path.exists(i): os.remove(i)

        else:
            if missing := [i for i in binaryData["files"] if not os.path.exists(i)]:
                print("\33[35m", "warning: missing binary file(s) detected: '", "', '".join(missing), "'.", "\33[0m", sep = "", flush = True)

    else:
        print("\33[35m", "warning: no binaries detected.", "\33[0m", sep = "", flush = True)

    if missing and int(os.environ.get("SDL_FIND_BINARIES", str(int(binaryData.get("find", True))))) > 0:
        binaryData["files"] += SDL_FIND_BINARIES(list(SDL_BINARY_VAR_MAP_INV.keys()))

    if missing and int(os.environ.get("SDL_DOWNLOAD_BINARIES", str(int(binaryData.get("repair", True))))) > 0:
        SDL_DOWNLOAD_BINARIES(binaryPath, SDL_SYSTEM, SDL_ARCH)

        with open(os.path.join(binaryPath, "metadata.json"), "r") as file:
            binaryData = json.load(file)
            binaryData["files"] = [absPath(i) for i in binaryData["files"]]

    for binary in SDL_BINARY_VAR_MAP_INV:
        for file in binaryData["files"].copy():
            if os.path.split(file)[1].split(".")[0].endswith(binary) and os.path.exists(file):
                binaryMap[binary], functions[binary] = ctypes.CDLL(os.path.abspath(file)), {}
                binaryData["files"].remove(file)

    if not int(os.environ.get("SDL_IGNORE_MISSING_BINARIES", "0")) > 0:
        if missing := [SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(binary) for binary in SDL_BINARY_VAR_MAP_INV if binary not in binaryMap]:
            print("\33[35m", "warning: missing binary file(s) detected: '", "', '".join(missing), "' (not repairable).", "\33[0m", sep = "", flush = True)

def SDL_ARRAY(*args: typing.List[typing.Any], **kwargs: typing.Dict[str, typing.Any]) -> typing.Tuple[ctypes.Array[typing.Any], int]:
    """Create a ctypes array."""
    return ((kwargs.get("type") or args[0].__class__) * len(args))(*args), len(args)

def SDL_DEREFERENCE(value: typing.Any) -> typing.Any:
    """Dereference a ctypes pointer or object."""
    if isinstance(value, ctypes._Pointer): return value.contents
    elif hasattr(value, "_obj"): return value._obj
    else: return value

def SDL_CACHE_FUNC(func: typing.Callable) -> typing.Callable:
    """Simple function cache decorator."""
    cache = {}

    def __inner__(*args: typing.List[typing.Any], **kwargs: typing.Dict[str, typing.Any]) -> typing.Any:
        _hash = hash((args, tuple(frozenset(sorted(kwargs.items())))))
        if _hash not in cache: cache.update({_hash: func(*args, **kwargs)})
        return cache.get(_hash, None)

    return __inner__

@SDL_CACHE_FUNC
def SDL_GET_BINARY_NAME(binary: typing.Any) -> str:
    """Get the name of an SDL3 binary."""
    return {v: k for k, v in __module__.binaryMap.items()}.get(binary, None)

def SDL_GET_BINARY(name: str) -> typing.Any:
    """Get an SDL3 binary by its name."""
    return __module__.binaryMap.get(name, None)

def SDL_SET_CURRENT_BINARY(name: str) -> None:
    """Set the current SDL3 binary by its name."""
    __module__.currentBinary = SDL_GET_BINARY(name)

def SDL_GET_CURRENT_BINARY() -> typing.Any:
    """Get the current SDL3 binary."""
    return __module__.currentBinary

def SDL_FUNC(name: str, retType: typing.Any, *args: typing.List[typing.Any]) -> None:
    """Define an SDL3 function."""

    if not (binary := SDL_GET_CURRENT_BINARY()):
        return

    if not hasattr(binary, name):
        if int(os.environ.get("SDL_IGNORE_MISSING_FUNCTIONS", "0")) > 0: return
        print("\33[35m", f"warning: function '{name}' not found in binary: '{SDL_GET_BINARY_NAME(binary)}'.", "\33[0m", sep = "", flush = True)

    else:
        func = getattr(binary, name)
        func.restype, func.argtypes = retType, args
        if not __doc_generator__: setattr(__module__, name, func)
        __module__.functions[SDL_GET_BINARY_NAME(binary)][name] = func

async def SDL_GET_LATEST_RELEASES() -> typing.Dict[str, str]:
    """Get latest releases of SDL3 modules from their official github repositories."""
    session, releases, tasks = aiohttp.ClientSession(), {}, []
    headers = {"Accept": "application/vnd.github+json"}

    if "SDL_GITHUB_TOKEN" in os.environ:
        headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

    for repo in SDL_REPOSITORIES:
        url = f"https://api.github.com/repos/libsdl-org/{repo}/releases"
        print("\33[32m", f"info: sending a request to '{url}'.", "\33[0m", sep = "", flush = True)
        tasks.append(asyncio.create_task(session.get(url, headers = headers, ssl = False)))

    responses = await asyncio.gather(*tasks)
    print("\33[32m", f"info: response gathering completed ({len(responses)} response(s)).", "\33[0m", sep = "", flush = True)

    for response, repo in zip(responses, SDL_REPOSITORIES):
        if response.status != 200:
            print("\33[35m", "warning: ", (await response.json())["message"].lower(), "\33[0m", sep = "", flush = True)
            if response.status == 403: await session.close(); return releases
            print("\33[35m", f"warning: failed to get latest releases of '{response.url}', skipping (status: {response.status}).", "\33[0m", sep = "", flush = True)
            releases[repo] = None

        else:
            latestRelease = (None, None)

            for release in await response.json():
                score = packaging.version.parse(release["tag_name"].split("-")[1])

                if not latestRelease[0] or score > latestRelease[0]:
                    latestRelease = (score, release["tag_name"])

            releases[repo] = latestRelease[1]

    await session.close()
    return releases

async def SDL_GET_FUNC_DESCRIPTIONS(funcs: typing.List[typing.Tuple[str, str]]) -> typing.Tuple[typing.List[str], typing.List[typing.List[str]]]:
    """Get descriptions and arguments of SDL3 functions from the official SDL3 wiki."""
    session, tasks = aiohttp.ClientSession(), []

    for module, func in funcs:
        url = f"https://wiki.libsdl.org/{module}/{func}"
        print("\33[32m", f"info: sending a request to '{url}'.", "\33[0m", sep = "", flush = True)
        tasks.append(asyncio.create_task(session.get(url, ssl = False)))
    
    responses = await asyncio.gather(*tasks)
    print("\33[32m", f"info: response gathering completed ({len(responses)} response(s)).", "\33[0m", sep = "", flush = True)
    descriptions, arguments = [], []

    for response in responses:
        if response.status != 200:
            print("\33[35m", f"warning: failed to get description of '{response.url}', skipping (status: {response.status}).", "\33[0m", sep = "", flush = True)
            descriptions.append(None)
            arguments.append(None)

        else:
            content = (await response.content.read()).decode()

            if "no such page" in content.lower():
                print("\33[35m", f"warning: no such page found for '{response.url}', skipping.", "\33[0m", sep = "", flush = True)
                descriptions.append(None)
                arguments.append(None)
                continue

            for a, b in zip(list(re.finditer(r"<p>", content)), list(re.finditer(r"</p>", content))):
                if content[a.end()] == "<": continue
                description = re.sub(r"<a href=\"([a-zA-Z0-9_]+)\">([a-zA-Z0-9_]+)</a>", rf"[\2]({'/'.join(str(response.url).split('/')[:-1])}/\1)", content[a.end():b.start()])
                descriptions.append(description.replace("<em>", "").replace("</em>", "").replace("<code>", "`").replace("</code>", "`").replace("<strong>", "").replace("</strong>", ""))
                break

            for a, b in zip(list(re.finditer(r"<code([a-zA-Z0-9_=#\s\"\-]*)>", content)), list(re.finditer(r"</code>", content))):
                if "sourceCode c" not in content[a.start():a.end()]: continue
                text = re.sub(r"<(/)?([a-zA-Z0-9_=#\s\"\-]+)>", r"", content[a.end():b.start()]).split(";")[0]
                if "\n" in text: text = " ".join(text.split("\n"))
                text = text[text.index("(") + 1:text.index(")") + 1].replace(")", ",").replace("[", "").replace("]", "")
                if text == "void,": text = ""
                arguments.append([text[arg.start():arg.end() - 1] for arg in re.finditer(r"([a-zA-Z0-9_]+)\,", text)])
                break

    await session.close()
    return descriptions, arguments

def SDL_GENERATE_DOCS() -> str:
    """Generate type hints and documentation for SDL3 functions."""

    __index, (descriptions, arguments) = -1, \
        asyncio.run(SDL_GET_FUNC_DESCRIPTIONS([(module, func) for module in __module__.functions for func in __module__.functions[module]]))

    for module in __module__.functions:
        for func in __module__.functions[module]:
            __module__.functions[module][func].__doc__ = \
                (descriptions[__index := __index + 1], arguments[__index])

    result = "\"\"\"This file is auto-generated.\"\"\"\n\nfrom .SDL import *\n\n"
    result += "from .__init__ import ctypes, typing, SDL_GET_BINARY, \\\n"
    result += f"{' ' * 4}{', '.join(list(SDL_BINARY_VAR_MAP.keys()))}\n\n"
    result += f"class POINTER:\n{' ' * 4}\"\"\"Simple pointer type.\"\"\"\n"
    result += f"{' ' * 4}@classmethod\n{' ' * 4}def __class_getitem__(cls, item):\n"
    result += f"{' ' * 8}return ctypes.POINTER(item)\n\n"
    types, definitions = set(), ""

    def SDL_GET_NAME(i):
        if i is None: return "None"
        if "CFunctionType" in i.__name__: return "ctypes._Pointer"
        if i.__name__.startswith("LP_"): types.add(i.__name__)
        if i.__name__.startswith("c_"): return f"ctypes.{i.__name__}"
        return i.__name__

    for index, module in enumerate(__module__.functions):
        if len(__module__.functions[module]) == 0: continue
        definitions += f"# {SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(module)} implementation.\n\n"

        for _index, func in enumerate(__module__.functions[module]):
            retType, argtypes, (description, arguments) = \
                __module__.functions[module][func].restype, __module__.functions[module][func].argtypes, __module__.functions[module][func].__doc__

            assert arguments is None or (arguments is not None and len(arguments) == len(argtypes)), f"argument count mismatch for 'https://wiki.libsdl.org/{module}/{func}'."
            arguments = [f"{'_' if arguments is None or arguments[i] in keyword.kwlist else ''}{i if arguments is None else arguments[i]}" for i in range(len(argtypes))]
            definitions += f"def {func}({', '.join([f'{arg}: {SDL_GET_NAME(type)}' for arg, type in zip(arguments, argtypes)])}) -> {SDL_GET_NAME(retType)}:\n"
            definitions += f"{' ' * 4}\"\"\"\n"
            if description is not None: definitions += f"    {description}\n"
            definitions += f"{' ' * 4}https://wiki.libsdl.org/{module}/{func}\n"
            definitions += f"{' ' * 4}\"\"\"\n"
            definitions += f"{' ' * 4}return SDL_GET_BINARY({SDL_BINARY_VAR_MAP_INV[module]}).{func}({', '.join(arguments)})"

            if _index != len(__module__.functions[module]) - 1:
                definitions += "\n\n"

        if index != len(__module__.functions) - 1 and len(list(__module__.functions.values())[index + 1]) != 0:
            definitions += "\n\n"

    for i in types:
        count, func = i.count("LP_"), i.replace("LP_", "")
        result += f"{i}: typing.TypeAlias = {'POINTER[' * count}{'ctypes.' if func.startswith('c_') else ''}{func}{']' * count}\n"

    return f"{result}\n{definitions}"

def SDL_GET_OR_GENERATE_DOCS() -> bytes:
    """Get type hints and documentation for SDL3 functions from github or generate it."""

    try:
        headers = {"Accept": "application/vnd.github+json"}

        if "SDL_GITHUB_TOKEN" in os.environ:
            headers["Authorization"] = f"Bearer {os.environ['SDL_GITHUB_TOKEN']}"

        for release in requests.get(f"https://api.github.com/repos/Aermoss/PySDL3/releases", headers = headers).json():
            if release["tag_name"] != f"v{__version__}":
                continue

            for asset in release["assets"]:
                if asset["name"] != f"{SDL_SYSTEM}-Docs.py":
                    continue

                with requests.get(asset["browser_download_url"], headers = headers, stream = True) as response:
                    assert response.status_code == 200, f"failed to get docs from github, status: {response.status_code}."
                    return bytearray().join([chunk for chunk in response.iter_content(chunk_size = 8192) if chunk])
            
    except requests.RequestException as exc:
        print("\33[31m", f"error: failed to get docs from github: {str(exc).lower()}.", "\33[0m", sep = "", flush = True)

    return SDL_GENERATE_DOCS().encode("utf-8")

if not __initialized__ and int(os.environ.get("SDL_CHECK_IMPORTS", "0")) > 0:
    existingModules = [i[:-3] for i in sorted(os.listdir(os.path.dirname(__file__))) if i.startswith("SDL_") and i.endswith(".py")]

    with open(os.path.join(os.path.dirname(__file__), "SDL.py"), "r") as file:
        importedModules = [i.replace("\n", "")[6:-9] for i in file.readlines() if i.startswith("from .") and "import *" in i]

    for index, module in enumerate(existingModules):
        if len(importedModules) <= index or module != importedModules[index]:
            print("\33[35m", f"warning: regenerating main module.", "\33[0m", sep = "", flush = True)

            with open(os.path.join(os.path.dirname(__file__), "SDL.py"), "w") as file:
                file.write("\n".join([f"from .{i} import *" for i in existingModules]))
                break

from sdl3.SDL import *

SDL_VERSIONNUM_STRING = lambda num: \
    f"{SDL_VERSIONNUM_MAJOR(num)}.{SDL_VERSIONNUM_MINOR(num)}.{SDL_VERSIONNUM_MICRO(num)}"

SDL_BINARY_VERSION, SDL_IMAGE_BINARY_VERSION, SDL_MIXER_BINARY_VERSION, SDL_TTF_BINARY_VERSION, SDL_RTF_BINARY_VERSION, SDL_NET_BINARY_VERSION = \
    (_ := SDL_GET_BINARY(SDL_BINARY)) and _.SDL_GetVersion(), (_ := SDL_GET_BINARY(SDL_IMAGE_BINARY)) and _.IMG_Version(), (_ := SDL_GET_BINARY(SDL_MIXER_BINARY)) and _.Mix_Version(), \
        (_ := SDL_GET_BINARY(SDL_TTF_BINARY)) and _.TTF_Version(), (_ := SDL_GET_BINARY(SDL_RTF_BINARY)) and _.RTF_Version(), (_ := SDL_GET_BINARY(SDL_NET_BINARY)) and _.SDLNet_Version()

if not __initialized__:
    if int(os.environ.get("SDL_CHECK_BINARY_VERSION", "1")) > 0:
        for i in SDL_BINARY_VAR_MAP:
            if (binaryVersion := locals()[f"{i}_VERSION"]) != (version := locals()[f"{i.replace('_BINARY', '')}_VERSION"]) and binaryVersion is not None:
                print("\33[35m", f"warning: version mismatch with binary: '{SDL_BINARY_PATTERNS[SDL_SYSTEM][0].format(SDL_BINARY_VAR_MAP[i])}' (expected: {SDL_VERSIONNUM_STRING(version)}, got: {SDL_VERSIONNUM_STRING(binaryVersion)}).", "\33[0m", sep = "", flush = True)

    if __doc_generator__:
        if not os.path.exists(__doc_file__):
            with open(__doc_file__, "wb") as file:
                file.write(SDL_GET_OR_GENERATE_DOCS())

        from .__doc__ import *

    else:
        if os.path.exists(__doc_file__):
            os.remove(__doc_file__)



from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_pixels import SDL_PixelFormat, SDL_Colorspace, SDL_Palette
from .SDL_properties import SDL_PropertiesID
from .SDL_blendmode import SDL_BlendMode
from .SDL_iostream import SDL_IOStream
from .SDL_rect import SDL_Rect

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_SurfaceFlags = ctypes.c_uint32

SDL_SURFACE_PREALLOCATED = 0x00000001
SDL_SURFACE_LOCK_NEEDED = 0x00000002
SDL_SURFACE_LOCKED = 0x00000004
SDL_SURFACE_SIMD_ALIGNED = 0x00000008

SDL_MUSTLOCK = lambda s: (s.flags & SDL_SURFACE_LOCK_NEEDED) == SDL_SURFACE_LOCK_NEEDED

SDL_ScaleMode = ctypes.c_uint32

SDL_SCALEMODE_NEAREST = 0
SDL_SCALEMODE_LINEAR = 1

SDL_FlipMode = ctypes.c_uint32

SDL_FLIP_NONE = 0
SDL_FLIP_HORIZONTAL = 1
SDL_FLIP_VERTICAL = 2

class SDL_Surface(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_SurfaceFlags),
        ("format", SDL_PixelFormat),
        ("w", ctypes.c_int),
        ("h", ctypes.c_int),
        ("pitch", ctypes.c_int),
        ("pixels", ctypes.c_void_p),
        ("refcount", ctypes.c_int),
        ("reserved", ctypes.c_void_p)
    ]

SDL_FUNC("SDL_CreateSurface", ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, SDL_PixelFormat)
SDL_FUNC("SDL_CreateSurfaceFrom", ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_DestroySurface", None, ctypes.POINTER(SDL_Surface))

SDL_FUNC("SDL_GetSurfaceProperties", SDL_PropertiesID, ctypes.POINTER(SDL_Surface))

SDL_PROP_SURFACE_SDR_WHITE_POINT_FLOAT = "SDL.surface.SDR_white_point".encode()
SDL_PROP_SURFACE_HDR_HEADROOM_FLOAT = "SDL.surface.HDR_headroom".encode()
SDL_PROP_SURFACE_TONEMAP_OPERATOR_STRING = "SDL.surface.tonemap".encode()

SDL_FUNC("SDL_SetSurfaceColorspace", ctypes.c_bool, ctypes.POINTER(SDL_Surface), SDL_Colorspace)
SDL_FUNC("SDL_GetSurfaceColorspace", SDL_Colorspace, ctypes.POINTER(SDL_Surface))

SDL_FUNC("SDL_CreateSurfacePalette", ctypes.POINTER(SDL_Palette), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_SetSurfacePalette", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Palette))
SDL_FUNC("SDL_GetSurfacePalette", ctypes.POINTER(SDL_Palette), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_AddSurfaceAlternateImage", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_SurfaceHasAlternateImages", ctypes.c_bool, ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_GetSurfaceImages", ctypes.POINTER(ctypes.POINTER(SDL_Surface)), ctypes.POINTER(SDL_Surface), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_RemoveSurfaceAlternateImages", None, ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_LockSurface", ctypes.c_bool, ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_UnlockSurface", None, ctypes.POINTER(SDL_Surface))

SDL_FUNC("SDL_LoadBMP_IO", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_IOStream), ctypes.c_bool)
SDL_FUNC("SDL_LoadBMP", ctypes.POINTER(SDL_Surface), ctypes.c_char_p)

SDL_FUNC("SDL_SaveBMP_IO", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_IOStream), ctypes.c_bool)
SDL_FUNC("SDL_SaveBMP", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_char_p)

SDL_FUNC("SDL_SetSurfaceRLE", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_bool)
SDL_FUNC("SDL_SurfaceHasRLE", ctypes.c_bool, ctypes.POINTER(SDL_Surface))

SDL_FUNC("SDL_SetSurfaceColorKey", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_bool, ctypes.c_uint32)
SDL_FUNC("SDL_SurfaceHasColorKey", ctypes.c_bool, ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_GetSurfaceColorKey", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(ctypes.c_uint32))

SDL_FUNC("SDL_SetSurfaceColorMod", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_GetSurfaceColorMod", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))

SDL_FUNC("SDL_SetSurfaceAlphaMod", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_uint8)
SDL_FUNC("SDL_GetSurfaceAlphaMod", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(ctypes.c_uint8))

SDL_FUNC("SDL_SetSurfaceBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Surface), SDL_BlendMode)
SDL_FUNC("SDL_GetSurfaceBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_BlendMode))

SDL_FUNC("SDL_SetSurfaceClipRect", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetSurfaceClipRect", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_FlipSurface", ctypes.c_bool, ctypes.POINTER(SDL_Surface), SDL_FlipMode)
SDL_FUNC("SDL_DuplicateSurface", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_ScaleSurface", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, SDL_ScaleMode)

SDL_FUNC("SDL_ConvertSurface", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Surface), SDL_PixelFormat)
SDL_FUNC("SDL_ConvertSurfaceAndColorspace", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Surface), SDL_PixelFormat, ctypes.POINTER(SDL_Palette), SDL_Colorspace, SDL_PropertiesID)
SDL_FUNC("SDL_ConvertPixels", ctypes.c_bool, ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_ConvertPixelsAndColorspace", ctypes.c_bool, ctypes.c_int, ctypes.c_int, SDL_PixelFormat, SDL_Colorspace, SDL_PropertiesID, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, SDL_Colorspace, SDL_PropertiesID, ctypes.c_void_p, ctypes.c_int)

SDL_FUNC("SDL_PremultiplyAlpha", ctypes.c_bool, ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, ctypes.c_bool)
SDL_FUNC("SDL_PremultiplySurfaceAlpha", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_bool)

SDL_FUNC("SDL_ClearSurface", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float)

SDL_FUNC("SDL_FillSurfaceRect", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.c_uint32)
SDL_FUNC("SDL_FillSurfaceRects", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.c_int, ctypes.c_uint32)

SDL_FUNC("SDL_BlitSurface", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_BlitSurfaceUnchecked", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_BlitSurfaceScaled", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), SDL_ScaleMode)
SDL_FUNC("SDL_BlitSurfaceUncheckedScaled", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), SDL_ScaleMode)
SDL_FUNC("SDL_StretchSurface", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), SDL_ScaleMode)
SDL_FUNC("SDL_BlitSurfaceTiled", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_BlitSurfaceTiledWithScale", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.c_float, SDL_ScaleMode, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_BlitSurface9Grid", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect), ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, SDL_ScaleMode, ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_MapSurfaceRGB", ctypes.c_uint32, ctypes.POINTER(SDL_Surface), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_MapSurfaceRGBA", ctypes.c_uint32, ctypes.POINTER(SDL_Surface), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)

SDL_FUNC("SDL_ReadSurfacePixel", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_ReadSurfacePixelFloat", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_WriteSurfacePixel", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_WriteSurfacePixelFloat", ctypes.c_bool, ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_MS_PER_SECOND = 1000
SDL_US_PER_SECOND = 1000000
SDL_NS_PER_SECOND = 1000000000
SDL_NS_PER_MS = 1000000
SDL_NS_PER_US = 1000

SDL_SECONDS_TO_NS = lambda s: s * SDL_NS_PER_SECOND
SDL_NS_TO_SECONDS = lambda ns: ns / SDL_NS_PER_SECOND
SDL_MS_TO_NS = lambda ms: ms * SDL_NS_PER_MS
SDL_NS_TO_MS = lambda ns: ns / SDL_NS_PER_MS
SDL_US_TO_NS = lambda us: us * SDL_NS_PER_US
SDL_NS_TO_US = lambda ns: ns / SDL_NS_PER_US

SDL_FUNC("SDL_GetTicks", ctypes.c_uint64)
SDL_FUNC("SDL_GetTicksNS", ctypes.c_uint64)
SDL_FUNC("SDL_GetPerformanceCounter", ctypes.c_uint64)
SDL_FUNC("SDL_GetPerformanceFrequency", ctypes.c_uint64)
SDL_FUNC("SDL_Delay", None, ctypes.c_uint32)
SDL_FUNC("SDL_DelayNS", None, ctypes.c_uint64)
SDL_FUNC("SDL_DelayPrecise", None, ctypes.c_uint64)

SDL_TimerID = ctypes.c_uint32
SDL_TimerCallback = ctypes.CFUNCTYPE(ctypes.c_uint32, ctypes.c_void_p, SDL_TimerID, ctypes.c_uint32)

SDL_FUNC("SDL_AddTimer", SDL_TimerID, ctypes.c_uint32, SDL_TimerCallback, ctypes.c_void_p)

SDL_NSTimerCallback = ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_void_p, SDL_TimerID, ctypes.c_uint64)

SDL_FUNC("SDL_AddTimerNS", SDL_TimerID, ctypes.c_uint64, SDL_NSTimerCallback, ctypes.c_void_p)
SDL_FUNC("SDL_RemoveTimer", ctypes.c_bool, SDL_TimerID)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_touch import SDL_TouchID
from .SDL_mouse import SDL_MouseID

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_PenID = ctypes.c_uint32

SDL_PEN_MOUSEID = SDL_MouseID(-2)
SDL_PEN_TOUCHID = SDL_TouchID(-2)

SDL_PenInputFlags = ctypes.c_uint32

SDL_PEN_INPUT_DOWN = 1 << 0
SDL_PEN_INPUT_BUTTON_1 = 1 << 1
SDL_PEN_INPUT_BUTTON_2 = 1 << 2
SDL_PEN_INPUT_BUTTON_3 = 1 << 3
SDL_PEN_INPUT_BUTTON_4 = 1 << 4
SDL_PEN_INPUT_BUTTON_5 = 1 << 5
SDL_PEN_INPUT_ERASER_TIP = 1 << 30

SDL_PenAxis = ctypes.c_int

SDL_PEN_AXIS_PRESSURE = 0
SDL_PEN_AXIS_XTILT = 1
SDL_PEN_AXIS_YTILT = 2
SDL_PEN_AXIS_DISTANCE = 3
SDL_PEN_AXIS_ROTATION = 4
SDL_PEN_AXIS_SLIDER = 5
SDL_PEN_AXIS_TANGENTIAL_PRESSURE = 6
SDL_PEN_AXIS_COUNT = 7


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_properties import SDL_PropertiesID
from .SDL_surface import SDL_FlipMode
from .SDL_video import SDL_Window
from .SDL_pixels import SDL_FColor
from .SDL_rect import SDL_Rect

SDL_SET_CURRENT_BINARY(SDL_BINARY)

class SDL_GPUDevice(ctypes.c_void_p):
    ...

class SDL_GPUBuffer(ctypes.c_void_p):
    ...

class SDL_GPUTransferBuffer(ctypes.c_void_p):
    ...

class SDL_GPUTexture(ctypes.c_void_p):
    ...

class SDL_GPUSampler(ctypes.c_void_p):
    ...

class SDL_GPUShader(ctypes.c_void_p):
    ...

class SDL_GPUComputePipeline(ctypes.c_void_p):
    ...

class SDL_GPUGraphicsPipeline(ctypes.c_void_p):
    ...

class SDL_GPUCommandBuffer(ctypes.c_void_p):
    ...

class SDL_GPURenderPass(ctypes.c_void_p):
    ...

class SDL_GPUComputePass(ctypes.c_void_p):
    ...

class SDL_GPUCopyPass(ctypes.c_void_p):
    ...

class SDL_GPUFence(ctypes.c_void_p):
    ...

SDL_GPUPrimitiveType = ctypes.c_int

SDL_GPU_PRIMITIVETYPE_TRIANGLELIST = 0
SDL_GPU_PRIMITIVETYPE_TRIANGLESTRIP = 1
SDL_GPU_PRIMITIVETYPE_LINELIST = 2
SDL_GPU_PRIMITIVETYPE_LINESTRIP = 3
SDL_GPU_PRIMITIVETYPE_POINTLIST = 4

SDL_GPULoadOp = ctypes.c_int

SDL_GPU_LOADOP_LOAD = 0
SDL_GPU_LOADOP_CLEAR = 1
SDL_GPU_LOADOP_DONT_CARE = 2

SDL_GPUStoreOp = ctypes.c_int

SDL_GPU_STOREOP_STORE = 0
SDL_GPU_STOREOP_DONT_CARE = 1
SDL_GPU_STOREOP_RESOLVE = 2
SDL_GPU_STOREOP_RESOLVE_AND_STORE = 3

SDL_GPUIndexElementSize = ctypes.c_int

SDL_GPU_INDEXELEMENTSIZE_16BIT = 0
SDL_GPU_INDEXELEMENTSIZE_32BIT = 1

SDL_GPUTextureFormat = ctypes.c_int

SDL_GPU_TEXTUREFORMAT_INVALID = 0

SDL_GPU_TEXTUREFORMAT_A8_UNORM = 1
SDL_GPU_TEXTUREFORMAT_R8_UNORM = 2
SDL_GPU_TEXTUREFORMAT_R8G8_UNORM = 3
SDL_GPU_TEXTUREFORMAT_R8G8B8A8_UNORM = 4
SDL_GPU_TEXTUREFORMAT_R16_UNORM = 5
SDL_GPU_TEXTUREFORMAT_R16G16_UNORM = 6
SDL_GPU_TEXTUREFORMAT_R16G16B16A16_UNORM = 7
SDL_GPU_TEXTUREFORMAT_R10G10B10A2_UNORM = 8
SDL_GPU_TEXTUREFORMAT_B5G6R5_UNORM = 9
SDL_GPU_TEXTUREFORMAT_B5G5R5A1_UNORM = 10
SDL_GPU_TEXTUREFORMAT_B4G4R4A4_UNORM = 11
SDL_GPU_TEXTUREFORMAT_B8G8R8A8_UNORM = 12
SDL_GPU_TEXTUREFORMAT_BC1_RGBA_UNORM = 13
SDL_GPU_TEXTUREFORMAT_BC2_RGBA_UNORM = 14
SDL_GPU_TEXTUREFORMAT_BC3_RGBA_UNORM = 15
SDL_GPU_TEXTUREFORMAT_BC4_R_UNORM = 16
SDL_GPU_TEXTUREFORMAT_BC5_RG_UNORM = 17
SDL_GPU_TEXTUREFORMAT_BC7_RGBA_UNORM = 18

SDL_GPU_TEXTUREFORMAT_BC6H_RGB_FLOAT = 19
SDL_GPU_TEXTUREFORMAT_BC6H_RGB_UFLOAT = 20

SDL_GPU_TEXTUREFORMAT_R8_SNORM = 21
SDL_GPU_TEXTUREFORMAT_R8G8_SNORM = 22
SDL_GPU_TEXTUREFORMAT_R8G8B8A8_SNORM = 23
SDL_GPU_TEXTUREFORMAT_R16_SNORM = 24
SDL_GPU_TEXTUREFORMAT_R16G16_SNORM = 25
SDL_GPU_TEXTUREFORMAT_R16G16B16A16_SNORM = 26

SDL_GPU_TEXTUREFORMAT_R16_FLOAT = 27
SDL_GPU_TEXTUREFORMAT_R16G16_FLOAT = 28
SDL_GPU_TEXTUREFORMAT_R16G16B16A16_FLOAT = 29
SDL_GPU_TEXTUREFORMAT_R32_FLOAT = 30
SDL_GPU_TEXTUREFORMAT_R32G32_FLOAT = 31
SDL_GPU_TEXTUREFORMAT_R32G32B32A32_FLOAT = 32

SDL_GPU_TEXTUREFORMAT_R11G11B10_UFLOAT = 33

SDL_GPU_TEXTUREFORMAT_R8_UINT = 34
SDL_GPU_TEXTUREFORMAT_R8G8_UINT = 35
SDL_GPU_TEXTUREFORMAT_R8G8B8A8_UINT = 36
SDL_GPU_TEXTUREFORMAT_R16_UINT = 37
SDL_GPU_TEXTUREFORMAT_R16G16_UINT = 38
SDL_GPU_TEXTUREFORMAT_R16G16B16A16_UINT = 39
SDL_GPU_TEXTUREFORMAT_R32_UINT = 40
SDL_GPU_TEXTUREFORMAT_R32G32_UINT = 41
SDL_GPU_TEXTUREFORMAT_R32G32B32A32_UINT = 42

SDL_GPU_TEXTUREFORMAT_R8_INT = 43
SDL_GPU_TEXTUREFORMAT_R8G8_INT = 44
SDL_GPU_TEXTUREFORMAT_R8G8B8A8_INT = 45
SDL_GPU_TEXTUREFORMAT_R16_INT = 46
SDL_GPU_TEXTUREFORMAT_R16G16_INT = 47
SDL_GPU_TEXTUREFORMAT_R16G16B16A16_INT = 48
SDL_GPU_TEXTUREFORMAT_R32_INT = 49
SDL_GPU_TEXTUREFORMAT_R32G32_INT = 50
SDL_GPU_TEXTUREFORMAT_R32G32B32A32_INT = 51

SDL_GPU_TEXTUREFORMAT_R8G8B8A8_UNORM_SRGB = 52
SDL_GPU_TEXTUREFORMAT_B8G8R8A8_UNORM_SRGB = 53

SDL_GPU_TEXTUREFORMAT_BC1_RGBA_UNORM_SRGB = 54
SDL_GPU_TEXTUREFORMAT_BC2_RGBA_UNORM_SRGB = 55
SDL_GPU_TEXTUREFORMAT_BC3_RGBA_UNORM_SRGB = 56
SDL_GPU_TEXTUREFORMAT_BC7_RGBA_UNORM_SRGB = 57

SDL_GPU_TEXTUREFORMAT_D16_UNORM = 58
SDL_GPU_TEXTUREFORMAT_D24_UNORM = 59
SDL_GPU_TEXTUREFORMAT_D32_FLOAT = 60
SDL_GPU_TEXTUREFORMAT_D24_UNORM_S8_UINT = 61
SDL_GPU_TEXTUREFORMAT_D32_FLOAT_S8_UINT = 62

SDL_GPU_TEXTUREFORMAT_ASTC_4x4_UNORM = 63
SDL_GPU_TEXTUREFORMAT_ASTC_5x4_UNORM = 64
SDL_GPU_TEXTUREFORMAT_ASTC_5x5_UNORM = 65
SDL_GPU_TEXTUREFORMAT_ASTC_6x5_UNORM = 66
SDL_GPU_TEXTUREFORMAT_ASTC_6x6_UNORM = 67
SDL_GPU_TEXTUREFORMAT_ASTC_8x5_UNORM = 68
SDL_GPU_TEXTUREFORMAT_ASTC_8x6_UNORM = 69
SDL_GPU_TEXTUREFORMAT_ASTC_8x8_UNORM = 70
SDL_GPU_TEXTUREFORMAT_ASTC_10x5_UNORM = 71
SDL_GPU_TEXTUREFORMAT_ASTC_10x6_UNORM = 72
SDL_GPU_TEXTUREFORMAT_ASTC_10x8_UNORM = 73
SDL_GPU_TEXTUREFORMAT_ASTC_10x10_UNORM = 74
SDL_GPU_TEXTUREFORMAT_ASTC_12x10_UNORM = 75
SDL_GPU_TEXTUREFORMAT_ASTC_12x12_UNORM = 76

SDL_GPU_TEXTUREFORMAT_ASTC_4x4_UNORM_SRGB = 77
SDL_GPU_TEXTUREFORMAT_ASTC_5x4_UNORM_SRGB = 78
SDL_GPU_TEXTUREFORMAT_ASTC_5x5_UNORM_SRGB = 79
SDL_GPU_TEXTUREFORMAT_ASTC_6x5_UNORM_SRGB = 80
SDL_GPU_TEXTUREFORMAT_ASTC_6x6_UNORM_SRGB = 81
SDL_GPU_TEXTUREFORMAT_ASTC_8x5_UNORM_SRGB = 82
SDL_GPU_TEXTUREFORMAT_ASTC_8x6_UNORM_SRGB = 83
SDL_GPU_TEXTUREFORMAT_ASTC_8x8_UNORM_SRGB = 84
SDL_GPU_TEXTUREFORMAT_ASTC_10x5_UNORM_SRGB = 85
SDL_GPU_TEXTUREFORMAT_ASTC_10x6_UNORM_SRGB = 86
SDL_GPU_TEXTUREFORMAT_ASTC_10x8_UNORM_SRGB = 87
SDL_GPU_TEXTUREFORMAT_ASTC_10x10_UNORM_SRGB = 88
SDL_GPU_TEXTUREFORMAT_ASTC_12x10_UNORM_SRGB = 89
SDL_GPU_TEXTUREFORMAT_ASTC_12x12_UNORM_SRGB = 90

SDL_GPU_TEXTUREFORMAT_ASTC_4x4_FLOAT = 91
SDL_GPU_TEXTUREFORMAT_ASTC_5x4_FLOAT = 92
SDL_GPU_TEXTUREFORMAT_ASTC_5x5_FLOAT = 93
SDL_GPU_TEXTUREFORMAT_ASTC_6x5_FLOAT = 94
SDL_GPU_TEXTUREFORMAT_ASTC_6x6_FLOAT = 95
SDL_GPU_TEXTUREFORMAT_ASTC_8x5_FLOAT = 96
SDL_GPU_TEXTUREFORMAT_ASTC_8x6_FLOAT = 97
SDL_GPU_TEXTUREFORMAT_ASTC_8x8_FLOAT = 98
SDL_GPU_TEXTUREFORMAT_ASTC_10x5_FLOAT = 99
SDL_GPU_TEXTUREFORMAT_ASTC_10x6_FLOAT = 100
SDL_GPU_TEXTUREFORMAT_ASTC_10x8_FLOAT = 101
SDL_GPU_TEXTUREFORMAT_ASTC_10x10_FLOAT = 102
SDL_GPU_TEXTUREFORMAT_ASTC_12x10_FLOAT = 103
SDL_GPU_TEXTUREFORMAT_ASTC_12x12_FLOAT = 104

SDL_GPUTextureUsageFlags = ctypes.c_uint32

SDL_GPU_TEXTUREUSAGE_SAMPLER = 1 << 0
SDL_GPU_TEXTUREUSAGE_COLOR_TARGET = 1 << 1
SDL_GPU_TEXTUREUSAGE_DEPTH_STENCIL_TARGET = 1 << 2
SDL_GPU_TEXTUREUSAGE_GRAPHICS_STORAGE_READ = 1 << 3
SDL_GPU_TEXTUREUSAGE_COMPUTE_STORAGE_READ = 1 << 4
SDL_GPU_TEXTUREUSAGE_COMPUTE_STORAGE_WRITE = 1 << 5
SDL_GPU_TEXTUREUSAGE_COMPUTE_STORAGE_SIMULTANEOUS_READ_WRITE = 1 << 6

SDL_GPUTextureType = ctypes.c_int

SDL_GPU_TEXTURETYPE_2D = 0
SDL_GPU_TEXTURETYPE_2D_ARRAY = 1
SDL_GPU_TEXTURETYPE_3D = 2
SDL_GPU_TEXTURETYPE_CUBE = 3
SDL_GPU_TEXTURETYPE_CUBE_ARRAY = 4

SDL_GPUSampleCount = ctypes.c_int

SDL_GPU_SAMPLECOUNT_1 = 0
SDL_GPU_SAMPLECOUNT_2 = 1
SDL_GPU_SAMPLECOUNT_4 = 2
SDL_GPU_SAMPLECOUNT_8 = 3

SDL_GPUCubeMapFace = ctypes.c_int

SDL_GPU_CUBEMAPFACE_POSITIVEX = 0
SDL_GPU_CUBEMAPFACE_NEGATIVEX = 1
SDL_GPU_CUBEMAPFACE_POSITIVEY = 2
SDL_GPU_CUBEMAPFACE_NEGATIVEY = 3
SDL_GPU_CUBEMAPFACE_POSITIVEZ = 4
SDL_GPU_CUBEMAPFACE_NEGATIVEZ = 5

SDL_GPUBufferUsageFlags = ctypes.c_uint32

SDL_GPU_BUFFERUSAGE_VERTEX = 1 << 0
SDL_GPU_BUFFERUSAGE_INDEX = 1 << 1
SDL_GPU_BUFFERUSAGE_INDIRECT = 1 << 2
SDL_GPU_BUFFERUSAGE_GRAPHICS_STORAGE_READ = 1 << 3
SDL_GPU_BUFFERUSAGE_COMPUTE_STORAGE_READ = 1 << 4
SDL_GPU_BUFFERUSAGE_COMPUTE_STORAGE_WRITE = 1 << 5

SDL_GPUTransferBufferUsage = ctypes.c_int

SDL_GPU_TRANSFERBUFFERUSAGE_UPLOAD = 0
SDL_GPU_TRANSFERBUFFERUSAGE_DOWNLOAD = 1

SDL_GPUShaderStage = ctypes.c_int

SDL_GPU_SHADERSTAGE_VERTEX = 0
SDL_GPU_SHADERSTAGE_FRAGMENT = 1

SDL_GPUShaderFormat = ctypes.c_uint32

SDL_GPU_SHADERFORMAT_INVALID = 0
SDL_GPU_SHADERFORMAT_PRIVATE = 1 << 0
SDL_GPU_SHADERFORMAT_SPIRV = 1 << 1
SDL_GPU_SHADERFORMAT_DXBC = 1 << 2
SDL_GPU_SHADERFORMAT_DXIL = 1 << 3
SDL_GPU_SHADERFORMAT_MSL = 1 << 4
SDL_GPU_SHADERFORMAT_METALLIB = 1 << 5

SDL_GPUVertexElementFormat = ctypes.c_int

SDL_GPU_VERTEXELEMENTFORMAT_INVALID = 0

SDL_GPU_VERTEXELEMENTFORMAT_INT = 1
SDL_GPU_VERTEXELEMENTFORMAT_INT2 = 2
SDL_GPU_VERTEXELEMENTFORMAT_INT3 = 3
SDL_GPU_VERTEXELEMENTFORMAT_INT4 = 4

SDL_GPU_VERTEXELEMENTFORMAT_UINT = 5
SDL_GPU_VERTEXELEMENTFORMAT_UINT2 = 6
SDL_GPU_VERTEXELEMENTFORMAT_UINT3 = 7
SDL_GPU_VERTEXELEMENTFORMAT_UINT4 = 8

SDL_GPU_VERTEXELEMENTFORMAT_FLOAT = 9
SDL_GPU_VERTEXELEMENTFORMAT_FLOAT2 = 10
SDL_GPU_VERTEXELEMENTFORMAT_FLOAT3 = 11
SDL_GPU_VERTEXELEMENTFORMAT_FLOAT4 = 12

SDL_GPU_VERTEXELEMENTFORMAT_BYTE2 = 13
SDL_GPU_VERTEXELEMENTFORMAT_BYTE4 = 14

SDL_GPU_VERTEXELEMENTFORMAT_UBYTE2 = 15
SDL_GPU_VERTEXELEMENTFORMAT_UBYTE4 = 16

SDL_GPU_VERTEXELEMENTFORMAT_BYTE2_NORM = 17
SDL_GPU_VERTEXELEMENTFORMAT_BYTE4_NORM = 18

SDL_GPU_VERTEXELEMENTFORMAT_UBYTE2_NORM = 19
SDL_GPU_VERTEXELEMENTFORMAT_UBYTE4_NORM = 20

SDL_GPU_VERTEXELEMENTFORMAT_SHORT2 = 21
SDL_GPU_VERTEXELEMENTFORMAT_SHORT4 = 22

SDL_GPU_VERTEXELEMENTFORMAT_USHORT2 = 23
SDL_GPU_VERTEXELEMENTFORMAT_USHORT4 = 24

SDL_GPU_VERTEXELEMENTFORMAT_SHORT2_NORM = 25
SDL_GPU_VERTEXELEMENTFORMAT_SHORT4_NORM = 26

SDL_GPU_VERTEXELEMENTFORMAT_USHORT2_NORM = 27
SDL_GPU_VERTEXELEMENTFORMAT_USHORT4_NORM = 28

SDL_GPU_VERTEXELEMENTFORMAT_HALF2 = 29
SDL_GPU_VERTEXELEMENTFORMAT_HALF4 = 30

SDL_GPUVertexInputRate = ctypes.c_int

SDL_GPU_VERTEXINPUTRATE_VERTEX = 0
SDL_GPU_VERTEXINPUTRATE_INSTANCE = 1

SDL_GPUFillMode = ctypes.c_int

SDL_GPU_FILLMODE_FILL = 0
SDL_GPU_FILLMODE_LINE = 1

SDL_GPUCullMode = ctypes.c_int

SDL_GPU_CULLMODE_NONE = 0
SDL_GPU_CULLMODE_FRONT = 1
SDL_GPU_CULLMODE_BACK = 2

SDL_GPUFrontFace = ctypes.c_int

SDL_GPU_FRONTFACE_COUNTER_CLOCKWISE = 0
SDL_GPU_FRONTFACE_CLOCKWISE = 1

SDL_GPUCompareOp = ctypes.c_int

SDL_GPU_COMPAREOP_INVALID = 0
SDL_GPU_COMPAREOP_NEVER = 1
SDL_GPU_COMPAREOP_LESS = 2
SDL_GPU_COMPAREOP_EQUAL = 3
SDL_GPU_COMPAREOP_LESS_OR_EQUAL = 4
SDL_GPU_COMPAREOP_GREATER = 5
SDL_GPU_COMPAREOP_NOT_EQUAL = 6
SDL_GPU_COMPAREOP_GREATER_OR_EQUAL = 7
SDL_GPU_COMPAREOP_ALWAYS = 8

SDL_GPUStencilOp = ctypes.c_int

SDL_GPU_STENCILOP_INVALID = 0
SDL_GPU_STENCILOP_KEEP = 1
SDL_GPU_STENCILOP_ZERO = 2
SDL_GPU_STENCILOP_REPLACE = 3
SDL_GPU_STENCILOP_INCREMENT_AND_CLAMP = 4
SDL_GPU_STENCILOP_DECREMENT_AND_CLAMP = 5
SDL_GPU_STENCILOP_INVERT = 6
SDL_GPU_STENCILOP_INCREMENT_AND_WRAP = 7
SDL_GPU_STENCILOP_DECREMENT_AND_WRAP = 8

SDL_GPUBlendOp = ctypes.c_int

SDL_GPU_BLENDOP_INVALID = 0
SDL_GPU_BLENDOP_ADD = 1
SDL_GPU_BLENDOP_SUBTRACT = 2
SDL_GPU_BLENDOP_REVERSE_SUBTRACT = 3
SDL_GPU_BLENDOP_MIN = 4
SDL_GPU_BLENDOP_MAX = 5

SDL_GPUBlendFactor = ctypes.c_int

SDL_GPU_BLENDFACTOR_INVALID = 0
SDL_GPU_BLENDFACTOR_ZERO = 1
SDL_GPU_BLENDFACTOR_ONE = 2
SDL_GPU_BLENDFACTOR_SRC_COLOR = 3
SDL_GPU_BLENDFACTOR_ONE_MINUS_SRC_COLOR = 4
SDL_GPU_BLENDFACTOR_DST_COLOR = 5
SDL_GPU_BLENDFACTOR_ONE_MINUS_DST_COLOR = 6
SDL_GPU_BLENDFACTOR_SRC_ALPHA = 7
SDL_GPU_BLENDFACTOR_ONE_MINUS_SRC_ALPHA = 8
SDL_GPU_BLENDFACTOR_DST_ALPHA = 9
SDL_GPU_BLENDFACTOR_ONE_MINUS_DST_ALPHA = 10
SDL_GPU_BLENDFACTOR_CONSTANT_COLOR = 11
SDL_GPU_BLENDFACTOR_ONE_MINUS_CONSTANT_COLOR = 12
SDL_GPU_BLENDFACTOR_SRC_ALPHA_SATURATE = 13

SDL_GPUColorComponentFlags = ctypes.c_uint8

SDL_GPU_COLORCOMPONENT_R = 1 << 0
SDL_GPU_COLORCOMPONENT_G = 1 << 1
SDL_GPU_COLORCOMPONENT_B = 1 << 2
SDL_GPU_COLORCOMPONENT_A = 1 << 3

SDL_GPUFilter = ctypes.c_int

SDL_GPU_FILTER_NEAREST = 0
SDL_GPU_FILTER_LINEAR = 1

SDL_GPUSamplerMipmapMode = ctypes.c_int

SDL_GPU_SAMPLERMIPMAPMODE_NEAREST = 0
SDL_GPU_SAMPLERMIPMAPMODE_LINEAR = 1

SDL_GPUSamplerAddressMode = ctypes.c_int

SDL_GPU_SAMPLERADDRESSMODE_REPEAT = 0
SDL_GPU_SAMPLERADDRESSMODE_MIRRORED_REPEAT = 1
SDL_GPU_SAMPLERADDRESSMODE_CLAMP_TO_EDGE = 2

SDL_GPUPresentMode = ctypes.c_int

SDL_GPU_PRESENTMODE_VSYNC = 0
SDL_GPU_PRESENTMODE_IMMEDIATE = 1
SDL_GPU_PRESENTMODE_MAILBOX = 2

SDL_GPUSwapchainComposition = ctypes.c_int

SDL_GPU_SWAPCHAINCOMPOSITION_SDR = 0
SDL_GPU_SWAPCHAINCOMPOSITION_SDR_LINEAR = 1
SDL_GPU_SWAPCHAINCOMPOSITION_HDR_EXTENDED_LINEAR = 2
SDL_GPU_SWAPCHAINCOMPOSITION_HDR10_ST2084 = 3

class SDL_GPUViewport(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("w", ctypes.c_float),
        ("h", ctypes.c_float),
        ("min_depth", ctypes.c_float),
        ("max_depth", ctypes.c_float)
    ]

class SDL_GPUTextureTransferInfo(ctypes.Structure):
    _fields_ = [
        ("transfer_buffer", ctypes.POINTER(SDL_GPUTransferBuffer)),
        ("offset", ctypes.c_uint32),
        ("pixels_per_row", ctypes.c_uint32),
        ("rows_per_layer", ctypes.c_uint32)
    ]

class SDL_GPUTransferBufferLocation(ctypes.Structure):
    _fields_ = [
        ("transfer_buffer", ctypes.POINTER(SDL_GPUTransferBuffer)),
        ("offset", ctypes.c_uint32)
    ]

class SDL_GPUTextureLocation(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("mip_level", ctypes.c_uint32),
        ("layer", ctypes.c_uint32),
        ("x", ctypes.c_uint32),
        ("y", ctypes.c_uint32),
        ("z", ctypes.c_uint32)
    ]

class SDL_GPUTextureRegion(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("mip_level", ctypes.c_uint32),
        ("layer", ctypes.c_uint32),
        ("x", ctypes.c_uint32),
        ("y", ctypes.c_uint32),
        ("z", ctypes.c_uint32),
        ("w", ctypes.c_uint32),
        ("h", ctypes.c_uint32),
        ("d", ctypes.c_uint32)
    ]

class SDL_GPUBlitRegion(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("mip_level", ctypes.c_uint32),
        ("layer_or_depth_plane", ctypes.c_uint32),
        ("x", ctypes.c_uint32),
        ("y", ctypes.c_uint32),
        ("w", ctypes.c_uint32),
        ("h", ctypes.c_uint32)
    ]

class SDL_GPUBufferLocation(ctypes.Structure):
    _fields_ = [
        ("buffer", ctypes.POINTER(SDL_GPUBuffer)),
        ("offset", ctypes.c_uint32)
    ]

class SDL_GPUBufferRegion(ctypes.Structure):
    _fields_ = [
        ("buffer", ctypes.POINTER(SDL_GPUBuffer)),
        ("offset", ctypes.c_uint32),
        ("size", ctypes.c_uint32)
    ]

class SDL_GPUIndirectDrawCommand(ctypes.Structure):
    _fields_ = [
        ("num_vertices", ctypes.c_uint32),
        ("num_instances", ctypes.c_uint32),
        ("first_vertex", ctypes.c_uint32),
        ("first_instance", ctypes.c_uint32)
    ]

class SDL_GPUIndexedIndirectDrawCommand(ctypes.Structure):
    _fields_ = [
        ("num_indices", ctypes.c_uint32),
        ("num_instances", ctypes.c_uint32),
        ("first_index", ctypes.c_uint32),
        ("vertex_offset", ctypes.c_int32),
        ("first_instance", ctypes.c_uint32)
    ]

class SDL_GPUIndirectDispatchCommand(ctypes.Structure):
    _fields_ = [
        ("groupcount_x", ctypes.c_uint32),
        ("groupcount_y", ctypes.c_uint32),
        ("groupcount_z", ctypes.c_uint32)
    ]

class SDL_GPUSamplerCreateInfo(ctypes.Structure):
    _fields_ = [
        ("min_filter", SDL_GPUFilter),
        ("mag_filter", SDL_GPUFilter),
        ("mipmap_mode", SDL_GPUSamplerMipmapMode),
        ("address_mode_u", SDL_GPUSamplerAddressMode),
        ("address_mode_v", SDL_GPUSamplerAddressMode),
        ("address_mode_w", SDL_GPUSamplerAddressMode),
        ("mip_lod_bias", ctypes.c_float),
        ("max_anisotropy", ctypes.c_float),
        ("compare_op", SDL_GPUCompareOp),
        ("min_lod", ctypes.c_float),
        ("max_lod", ctypes.c_float),
        ("enable_anisotropy", ctypes.c_bool),
        ("enable_compare", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUVertexBufferDescription(ctypes.Structure):
    _fields_ = [
        ("slot", ctypes.c_uint32),
        ("pitch", ctypes.c_uint32),
        ("input_rate", SDL_GPUVertexInputRate),
        ("instance_step_rate", ctypes.c_uint32)
    ]

class SDL_GPUVertexAttribute(ctypes.Structure):
    _fields_ = [
        ("location", ctypes.c_uint32),
        ("buffer_slot", ctypes.c_uint32),
        ("format", SDL_GPUVertexElementFormat),
        ("offset", ctypes.c_uint32)
    ]

class SDL_GPUVertexInputState(ctypes.Structure):
    _fields_ = [
        ("vertex_buffer_descriptions", ctypes.POINTER(SDL_GPUVertexBufferDescription)),
        ("num_vertex_buffers", ctypes.c_uint32),
        ("vertex_attributes", ctypes.POINTER(SDL_GPUVertexAttribute)),
        ("num_vertex_attributes", ctypes.c_uint32)
    ]

class SDL_GPUStencilOpState(ctypes.Structure):
    _fields_ = [
        ("fail_op", SDL_GPUStencilOp),
        ("pass_op", SDL_GPUStencilOp),
        ("depth_fail_op", SDL_GPUStencilOp),
        ("compare_op", SDL_GPUCompareOp)
    ]

class SDL_GPUColorTargetBlendState(ctypes.Structure):
    _fields_ = [
        ("src_color_blendfactor", SDL_GPUBlendFactor),
        ("dst_color_blendfactor", SDL_GPUBlendFactor),
        ("color_blend_op", SDL_GPUBlendOp),
        ("src_alpha_blendfactor", SDL_GPUBlendFactor),
        ("dst_alpha_blendfactor", SDL_GPUBlendFactor),
        ("alpha_blend_op", SDL_GPUBlendOp),
        ("color_write_mask", SDL_GPUColorComponentFlags),
        ("enable_blend", ctypes.c_bool),
        ("enable_color_write_mask", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8)
    ]

class SDL_GPUShaderCreateInfo(ctypes.Structure):
    _fields_ = [
        ("code_size", ctypes.c_size_t),
        ("code", ctypes.POINTER(ctypes.c_uint8)),
        ("entrypoint", ctypes.c_char_p),
        ("format", SDL_GPUShaderFormat),
        ("stage", SDL_GPUShaderStage),
        ("num_samplers", ctypes.c_uint32),
        ("num_storage_textures", ctypes.c_uint32),
        ("num_storage_buffers", ctypes.c_uint32),
        ("num_uniform_buffers", ctypes.c_uint32),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUTextureCreateInfo(ctypes.Structure):
    _fields_ = [
        ("type", SDL_GPUTextureType),
        ("format", SDL_GPUTextureFormat),
        ("usage", SDL_GPUTextureUsageFlags),
        ("width", ctypes.c_uint32),
        ("height", ctypes.c_uint32),
        ("layer_count_or_depth", ctypes.c_uint32),
        ("num_levels", ctypes.c_uint32),
        ("sample_count", SDL_GPUSampleCount),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUBufferCreateInfo(ctypes.Structure):
    _fields_ = [
        ("usage", SDL_GPUBufferUsageFlags),
        ("size", ctypes.c_uint32),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUTransferBufferCreateInfo(ctypes.Structure):
    _fields_ = [
        ("usage", SDL_GPUTransferBufferUsage),
        ("size", ctypes.c_uint32),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPURasterizerState(ctypes.Structure):
    _fields_ = [
        ("fill_mode", SDL_GPUFillMode),
        ("cull_mode", SDL_GPUCullMode),
        ("front_face", SDL_GPUFrontFace),
        ("depth_bias_constant_factor", ctypes.c_float),
        ("depth_bias_clamp", ctypes.c_float),
        ("depth_bias_slope_factor", ctypes.c_float),
        ("enable_depth_bias", ctypes.c_bool),
        ("enable_depth_clip", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8)
    ]

class SDL_GPUMultisampleState(ctypes.Structure):
    _fields_ = [
        ("sample_count", SDL_GPUSampleCount),
        ("sample_mask", ctypes.c_uint32),
        ("enable_mask", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

class SDL_GPUDepthStencilState(ctypes.Structure):
    _fields_ = [
        ("compare_op", SDL_GPUCompareOp),
        ("back_stencil_state", SDL_GPUStencilOpState),
        ("front_stencil_state", SDL_GPUStencilOpState),
        ("compare_mask", ctypes.c_uint8),
        ("write_mask", ctypes.c_uint8),
        ("enable_depth_test", ctypes.c_bool),
        ("enable_depth_write", ctypes.c_bool),
        ("enable_stencil_test", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

class SDL_GPUColorTargetDescription(ctypes.Structure):
    _fields_ = [
        ("format", SDL_GPUTextureFormat),
        ("blend_state", SDL_GPUColorTargetBlendState)
    ]

class SDL_GPUGraphicsPipelineTargetInfo(ctypes.Structure):
    _fields_ = [
        ("color_target_descriptions", ctypes.POINTER(SDL_GPUColorTargetDescription)),
        ("num_color_targets", ctypes.c_uint32),
        ("depth_stencil_format", SDL_GPUTextureFormat),
        ("has_depth_stencil_target", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

class SDL_GPUGraphicsPipelineCreateInfo(ctypes.Structure):
    _fields_ = [
        ("vertex_shader", ctypes.POINTER(SDL_GPUShader)),
        ("fragment_shader", ctypes.POINTER(SDL_GPUShader)),
        ("vertex_input_state", SDL_GPUVertexInputState),
        ("primitive_type", SDL_GPUPrimitiveType),
        ("rasterizer_state", SDL_GPURasterizerState),
        ("multisample_state", SDL_GPUMultisampleState),
        ("depth_stencil_state", SDL_GPUDepthStencilState),
        ("target_info", SDL_GPUGraphicsPipelineTargetInfo),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUComputePipelineCreateInfo(ctypes.Structure):
    _fields_ = [
        ("code_size", ctypes.c_size_t),
        ("code", ctypes.POINTER(ctypes.c_uint8)),
        ("entrypoint", ctypes.c_char_p),
        ("format", SDL_GPUShaderFormat),
        ("num_samplers", ctypes.c_uint32),
        ("num_readonly_storage_textures", ctypes.c_uint32),
        ("num_readonly_storage_buffers", ctypes.c_uint32),
        ("num_readwrite_storage_textures", ctypes.c_uint32),
        ("num_readwrite_storage_buffers", ctypes.c_uint32),
        ("num_uniform_buffers", ctypes.c_uint32),
        ("threadcount_x", ctypes.c_uint32),
        ("threadcount_y", ctypes.c_uint32),
        ("threadcount_z", ctypes.c_uint32),
        ("props", SDL_PropertiesID)
    ]

class SDL_GPUColorTargetInfo(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("mip_level", ctypes.c_uint32),
        ("layer_or_depth_plane", ctypes.c_uint32),
        ("clear_color", SDL_FColor),
        ("load_op", SDL_GPULoadOp),
        ("store_op", SDL_GPUStoreOp),
        ("resolve_texture", ctypes.POINTER(SDL_GPUTexture)),
        ("resolve_mip_level", ctypes.c_uint32),
        ("resolve_layer", ctypes.c_uint32),
        ("cycle", ctypes.c_bool),
        ("cycle_resolve_texture", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8)
    ]

class SDL_GPUDepthStencilTargetInfo(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("clear_depth", ctypes.c_float),
        ("load_op", SDL_GPULoadOp),
        ("store_op", SDL_GPUStoreOp),
        ("stencil_load_op", SDL_GPULoadOp),
        ("stencil_store_op", SDL_GPUStoreOp),
        ("cycle", ctypes.c_bool),
        ("clear_stencil", ctypes.c_uint8),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8)
    ]

class SDL_GPUBlitInfo(ctypes.Structure):
    _fields_ = [
        ("source", SDL_GPUBlitRegion),
        ("destination", SDL_GPUBlitRegion),
        ("load_op", SDL_GPULoadOp),
        ("clear_color", SDL_FColor),
        ("flip_mode", SDL_FlipMode),
        ("filter", SDL_GPUFilter),
        ("cycle", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

class SDL_GPUBufferBinding(ctypes.Structure):
    _fields_ = [
        ("buffer", ctypes.POINTER(SDL_GPUBuffer)),
        ("offset", ctypes.c_uint32)
    ]

class SDL_GPUTextureSamplerBinding(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("sampler", ctypes.POINTER(SDL_GPUSampler))
    ]

class SDL_GPUStorageBufferReadWriteBinding(ctypes.Structure):
    _fields_ = [
        ("buffer", ctypes.POINTER(SDL_GPUBuffer)),
        ("cycle", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

class SDL_GPUStorageTextureReadWriteBinding(ctypes.Structure):
    _fields_ = [
        ("texture", ctypes.POINTER(SDL_GPUTexture)),
        ("mip_level", ctypes.c_uint32),
        ("layer", ctypes.c_uint32),
        ("cycle", ctypes.c_bool),
        ("padding1", ctypes.c_uint8),
        ("padding2", ctypes.c_uint8),
        ("padding3", ctypes.c_uint8)
    ]

SDL_FUNC("SDL_GPUSupportsShaderFormats", ctypes.c_bool, SDL_GPUShaderFormat, ctypes.c_char_p)
SDL_FUNC("SDL_GPUSupportsProperties", ctypes.c_bool, SDL_PropertiesID)

SDL_FUNC("SDL_CreateGPUDevice", ctypes.POINTER(SDL_GPUDevice), SDL_GPUShaderFormat, ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("SDL_CreateGPUDeviceWithProperties", ctypes.POINTER(SDL_GPUDevice), SDL_PropertiesID)

SDL_PROP_GPU_DEVICE_CREATE_DEBUGMODE_BOOLEAN = "SDL.gpu.device.create.debugmode".encode()
SDL_PROP_GPU_DEVICE_CREATE_PREFERLOWPOWER_BOOLEAN = "SDL.gpu.device.create.preferlowpower".encode()
SDL_PROP_GPU_DEVICE_CREATE_NAME_STRING = "SDL.gpu.device.create.name".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_PRIVATE_BOOLEAN = "SDL.gpu.device.create.shaders.private".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_SPIRV_BOOLEAN = "SDL.gpu.device.create.shaders.spirv".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_DXBC_BOOLEAN = "SDL.gpu.device.create.shaders.dxbc".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_DXIL_BOOLEAN = "SDL.gpu.device.create.shaders.dxil".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_MSL_BOOLEAN = "SDL.gpu.device.create.shaders.msl".encode()
SDL_PROP_GPU_DEVICE_CREATE_SHADERS_METALLIB_BOOLEAN = "SDL.gpu.device.create.shaders.metallib".encode()
SDL_PROP_GPU_DEVICE_CREATE_D3D12_SEMANTIC_NAME_STRING = "SDL.gpu.device.create.d3d12.semantic".encode()

SDL_FUNC("SDL_DestroyGPUDevice", None, ctypes.POINTER(SDL_GPUDevice))

SDL_FUNC("SDL_GetNumGPUDrivers", ctypes.c_int)
SDL_FUNC("SDL_GetGPUDriver", ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_GetGPUDeviceDriver", ctypes.c_char_p, ctypes.POINTER(SDL_GPUDevice))
SDL_FUNC("SDL_GetGPUShaderFormats", SDL_GPUShaderFormat, ctypes.POINTER(SDL_GPUDevice))

SDL_FUNC("SDL_CreateGPUComputePipeline", ctypes.POINTER(SDL_GPUComputePipeline), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUComputePipelineCreateInfo))

SDL_PROP_GPU_COMPUTEPIPELINE_CREATE_NAME_STRING = "SDL.gpu.computepipeline.create.name".encode()

SDL_FUNC("SDL_CreateGPUGraphicsPipeline", ctypes.POINTER(SDL_GPUGraphicsPipeline), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUGraphicsPipelineCreateInfo))

SDL_PROP_GPU_GRAPHICSPIPELINE_CREATE_NAME_STRING = "SDL.gpu.graphicspipeline.create.name".encode()

SDL_FUNC("SDL_CreateGPUSampler", ctypes.POINTER(SDL_GPUSampler), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUSamplerCreateInfo))

SDL_PROP_GPU_SAMPLER_CREATE_NAME_STRING = "SDL.gpu.sampler.create.name".encode()

SDL_FUNC("SDL_CreateGPUShader", ctypes.POINTER(SDL_GPUShader), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUShaderCreateInfo))

SDL_PROP_GPU_SHADER_CREATE_NAME_STRING = "SDL.gpu.shader.create.name".encode()

SDL_FUNC("SDL_CreateGPUTexture", ctypes.POINTER(SDL_GPUTexture), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTextureCreateInfo))

SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_R_FLOAT = "SDL.gpu.texture.create.d3d12.clear.r".encode()
SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_G_FLOAT = "SDL.gpu.texture.create.d3d12.clear.g".encode()
SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_B_FLOAT = "SDL.gpu.texture.create.d3d12.clear.b".encode()
SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_A_FLOAT = "SDL.gpu.texture.create.d3d12.clear.a".encode()
SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_DEPTH_FLOAT = "SDL.gpu.texture.create.d3d12.clear.depth".encode()
SDL_PROP_GPU_TEXTURE_CREATE_D3D12_CLEAR_STENCIL_UINT8 = "SDL.gpu.texture.create.d3d12.clear.stencil".encode()
SDL_PROP_GPU_TEXTURE_CREATE_NAME_STRING = "SDL.gpu.texture.create.name".encode()

SDL_FUNC("SDL_CreateGPUBuffer", ctypes.POINTER(SDL_GPUBuffer), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUBufferCreateInfo))

SDL_PROP_GPU_BUFFER_CREATE_NAME_STRING = "SDL.gpu.buffer.create.name".encode()

SDL_FUNC("SDL_CreateGPUTransferBuffer", ctypes.POINTER(SDL_GPUTransferBuffer), ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTransferBufferCreateInfo))

SDL_PROP_GPU_TRANSFERBUFFER_CREATE_NAME_STRING = "SDL.gpu.transferbuffer.create.name".encode()

SDL_FUNC("SDL_SetGPUBufferName", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUBuffer), ctypes.c_char_p)
SDL_FUNC("SDL_SetGPUTextureName", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTexture), ctypes.c_char_p)

SDL_FUNC("SDL_InsertGPUDebugLabel", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.c_char_p)
SDL_FUNC("SDL_PushGPUDebugGroup", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.c_char_p)
SDL_FUNC("SDL_PopGPUDebugGroup", None, ctypes.POINTER(SDL_GPUCommandBuffer))

SDL_FUNC("SDL_ReleaseGPUTexture", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTexture))
SDL_FUNC("SDL_ReleaseGPUSampler", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUSampler))
SDL_FUNC("SDL_ReleaseGPUBuffer", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUBuffer))
SDL_FUNC("SDL_ReleaseGPUTransferBuffer", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTransferBuffer))
SDL_FUNC("SDL_ReleaseGPUComputePipeline", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUComputePipeline))
SDL_FUNC("SDL_ReleaseGPUShader", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUShader))
SDL_FUNC("SDL_ReleaseGPUGraphicsPipeline", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUGraphicsPipeline))

SDL_FUNC("SDL_AcquireGPUCommandBuffer", ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_GPUDevice))

SDL_FUNC("SDL_PushGPUVertexUniformData", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.c_uint32, ctypes.c_void_p, ctypes.c_uint32)
SDL_FUNC("SDL_PushGPUFragmentUniformData", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.c_uint32, ctypes.c_void_p, ctypes.c_uint32)
SDL_FUNC("SDL_PushGPUComputeUniformData", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.c_uint32, ctypes.c_void_p, ctypes.c_uint32)

SDL_FUNC("SDL_BeginGPURenderPass", ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_GPUColorTargetInfo), ctypes.c_uint32, ctypes.POINTER(SDL_GPUDepthStencilTargetInfo))
SDL_FUNC("SDL_BindGPUGraphicsPipeline", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUGraphicsPipeline))
SDL_FUNC("SDL_SetGPUViewport", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUViewport))
SDL_FUNC("SDL_SetGPUScissor", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_SetGPUBlendConstants", None, ctypes.POINTER(SDL_GPURenderPass), SDL_FColor)
SDL_FUNC("SDL_SetGPUStencilReference", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint8)
SDL_FUNC("SDL_BindGPUVertexBuffers", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUBufferBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUIndexBuffer", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUBufferBinding), SDL_GPUIndexElementSize)
SDL_FUNC("SDL_BindGPUVertexSamplers", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUTextureSamplerBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUVertexStorageTextures", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(ctypes.POINTER(SDL_GPUTexture)), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUVertexStorageBuffers", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUBufferBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUFragmentSamplers", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUTextureSamplerBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUFragmentStorageTextures", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(ctypes.POINTER(SDL_GPUTexture)), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUFragmentStorageBuffers", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUBufferBinding), ctypes.c_uint32)
SDL_FUNC("SDL_DrawGPUIndexedPrimitives", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_int32, ctypes.c_uint32)
SDL_FUNC("SDL_DrawGPUPrimitives", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32)
SDL_FUNC("SDL_DrawGPUPrimitivesIndirect", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUBuffer), ctypes.c_uint32, ctypes.c_uint32)
SDL_FUNC("SDL_DrawGPUIndexedPrimitivesIndirect", None, ctypes.POINTER(SDL_GPURenderPass), ctypes.POINTER(SDL_GPUBuffer), ctypes.c_uint32, ctypes.c_uint32)
SDL_FUNC("SDL_EndGPURenderPass", None, ctypes.POINTER(SDL_GPURenderPass))

SDL_FUNC("SDL_BeginGPUComputePass", ctypes.POINTER(SDL_GPUComputePass), ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_GPUStorageTextureReadWriteBinding), ctypes.c_uint32, ctypes.POINTER(SDL_GPUStorageBufferReadWriteBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUComputePipeline", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.POINTER(SDL_GPUComputePipeline))
SDL_FUNC("SDL_BindGPUComputeSamplers", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUTextureSamplerBinding), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUComputeStorageTextures", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.c_uint32, ctypes.POINTER(ctypes.POINTER(SDL_GPUTexture)), ctypes.c_uint32)
SDL_FUNC("SDL_BindGPUComputeStorageBuffers", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.c_uint32, ctypes.POINTER(SDL_GPUBufferBinding), ctypes.c_uint32)
SDL_FUNC("SDL_DispatchGPUCompute", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32)
SDL_FUNC("SDL_DispatchGPUComputeIndirect", None, ctypes.POINTER(SDL_GPUComputePass), ctypes.POINTER(SDL_GPUBuffer), ctypes.c_uint32)
SDL_FUNC("SDL_EndGPUComputePass", None, ctypes.POINTER(SDL_GPUComputePass))

SDL_FUNC("SDL_MapGPUTransferBuffer", ctypes.c_void_p, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTransferBuffer), ctypes.c_bool)
SDL_FUNC("SDL_UnmapGPUTransferBuffer", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUTransferBuffer))

SDL_FUNC("SDL_BeginGPUCopyPass", ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUCommandBuffer))
SDL_FUNC("SDL_UploadToGPUTexture", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUTextureTransferInfo), ctypes.POINTER(SDL_GPUTextureRegion), ctypes.c_bool)
SDL_FUNC("SDL_UploadToGPUBuffer", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUTransferBufferLocation), ctypes.POINTER(SDL_GPUBufferRegion), ctypes.c_bool)
SDL_FUNC("SDL_CopyGPUTextureToTexture", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUTextureLocation), ctypes.POINTER(SDL_GPUTextureLocation), ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_bool)
SDL_FUNC("SDL_CopyGPUBufferToBuffer", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUBufferLocation), ctypes.POINTER(SDL_GPUBufferLocation), ctypes.c_uint32, ctypes.c_bool)
SDL_FUNC("SDL_DownloadFromGPUTexture", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUTextureRegion), ctypes.POINTER(SDL_GPUTextureTransferInfo))
SDL_FUNC("SDL_DownloadFromGPUBuffer", None, ctypes.POINTER(SDL_GPUCopyPass), ctypes.POINTER(SDL_GPUBufferRegion), ctypes.POINTER(SDL_GPUTransferBufferLocation))
SDL_FUNC("SDL_EndGPUCopyPass", None, ctypes.POINTER(SDL_GPUCopyPass))

SDL_FUNC("SDL_GenerateMipmapsForGPUTexture", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_GPUTexture))
SDL_FUNC("SDL_BlitGPUTexture", None, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_GPUBlitInfo))

SDL_FUNC("SDL_WindowSupportsGPUSwapchainComposition", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window), SDL_GPUSwapchainComposition)
SDL_FUNC("SDL_WindowSupportsGPUPresentMode", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window), SDL_GPUPresentMode)
SDL_FUNC("SDL_ClaimWindowForGPUDevice", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_ReleaseWindowFromGPUDevice", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_SetGPUSwapchainParameters", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window), SDL_GPUSwapchainComposition, SDL_GPUPresentMode)
SDL_FUNC("SDL_SetGPUAllowedFramesInFlight", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.c_uint32)
SDL_FUNC("SDL_GetGPUSwapchainTextureFormat", SDL_GPUTextureFormat, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_AcquireGPUSwapchainTexture", ctypes.c_bool, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_Window), ctypes.POINTER(ctypes.POINTER(SDL_GPUTexture)), ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_WaitForGPUSwapchain", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_WaitAndAcquireGPUSwapchainTexture", ctypes.c_bool, ctypes.POINTER(SDL_GPUCommandBuffer), ctypes.POINTER(SDL_Window), ctypes.POINTER(ctypes.POINTER(SDL_GPUTexture)), ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint32))

SDL_FUNC("SDL_SubmitGPUCommandBuffer", ctypes.c_bool, ctypes.POINTER(SDL_GPUCommandBuffer))
SDL_FUNC("SDL_SubmitGPUCommandBufferAndAcquireFence", ctypes.POINTER(SDL_GPUFence), ctypes.POINTER(SDL_GPUCommandBuffer))
SDL_FUNC("SDL_CancelGPUCommandBuffer", ctypes.c_bool, ctypes.POINTER(SDL_GPUCommandBuffer))

SDL_FUNC("SDL_WaitForGPUIdle", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice))
SDL_FUNC("SDL_WaitForGPUFences", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.c_bool, ctypes.POINTER(ctypes.POINTER(SDL_GPUFence)), ctypes.c_uint32)

SDL_FUNC("SDL_QueryGPUFence", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUFence))
SDL_FUNC("SDL_ReleaseGPUFence", None, ctypes.POINTER(SDL_GPUDevice), ctypes.POINTER(SDL_GPUFence))

SDL_FUNC("SDL_GPUTextureFormatTexelBlockSize", ctypes.c_uint32, SDL_GPUTextureFormat)
SDL_FUNC("SDL_GPUTextureSupportsFormat", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), SDL_GPUTextureFormat, SDL_GPUTextureType, SDL_GPUTextureUsageFlags)
SDL_FUNC("SDL_GPUTextureSupportsSampleCount", ctypes.c_bool, ctypes.POINTER(SDL_GPUDevice), SDL_GPUTextureFormat, SDL_GPUSampleCount)
SDL_FUNC("SDL_CalculateGPUTextureFormatSize", ctypes.c_uint32, SDL_GPUTextureFormat, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_video import SDL_Window
from .SDL_surface import SDL_Surface

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_MouseID = ctypes.c_uint32

class SDL_Cursor(ctypes.c_void_p):
    ...

SDL_SystemCursor = ctypes.c_int

SDL_SYSTEM_CURSOR_DEFAULT = 0
SDL_SYSTEM_CURSOR_TEXT = 1
SDL_SYSTEM_CURSOR_WAIT = 2
SDL_SYSTEM_CURSOR_CROSSHAIR = 3
SDL_SYSTEM_CURSOR_PROGRESS = 4
SDL_SYSTEM_CURSOR_NWSE_RESIZE = 5
SDL_SYSTEM_CURSOR_NESW_RESIZE = 6
SDL_SYSTEM_CURSOR_EW_RESIZE = 7
SDL_SYSTEM_CURSOR_NS_RESIZE = 8
SDL_SYSTEM_CURSOR_MOVE = 9
SDL_SYSTEM_CURSOR_NOT_ALLOWED = 10
SDL_SYSTEM_CURSOR_POINTER = 11
SDL_SYSTEM_CURSOR_NW_RESIZE = 12
SDL_SYSTEM_CURSOR_N_RESIZE = 13
SDL_SYSTEM_CURSOR_NE_RESIZE = 14
SDL_SYSTEM_CURSOR_E_RESIZE = 15
SDL_SYSTEM_CURSOR_SE_RESIZE = 16
SDL_SYSTEM_CURSOR_S_RESIZE = 17
SDL_SYSTEM_CURSOR_SW_RESIZE = 18
SDL_SYSTEM_CURSOR_W_RESIZE = 19
SDL_SYSTEM_CURSOR_COUNT = 20

SDL_MouseWheelDirection = ctypes.c_int

SDL_MOUSEWHEEL_NORMAL = 0
SDL_MOUSEWHEEL_FLIPPED = 1

SDL_MouseButtonFlags = ctypes.c_uint32

SDL_BUTTON_LEFT = 1
SDL_BUTTON_MIDDLE = 2
SDL_BUTTON_RIGHT = 3
SDL_BUTTON_X1 = 4
SDL_BUTTON_X2 = 5

SDL_BUTTON_MASK = lambda x: 1 << (x - 1)
SDL_BUTTON_LMASK = SDL_BUTTON_MASK(SDL_BUTTON_LEFT)
SDL_BUTTON_MMASK = SDL_BUTTON_MASK(SDL_BUTTON_MIDDLE)
SDL_BUTTON_RMASK = SDL_BUTTON_MASK(SDL_BUTTON_RIGHT)
SDL_BUTTON_X1MASK = SDL_BUTTON_MASK(SDL_BUTTON_X1)
SDL_BUTTON_X2MASK = SDL_BUTTON_MASK(SDL_BUTTON_X2)

SDL_FUNC("SDL_HasMouse", ctypes.c_bool)
SDL_FUNC("SDL_GetMice", ctypes.POINTER(SDL_MouseID), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetMouseNameForID", ctypes.c_char_p, SDL_MouseID)
SDL_FUNC("SDL_GetMouseFocus", ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_GetMouseState", SDL_MouseButtonFlags, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))
SDL_FUNC("SDL_GetGlobalMouseState", SDL_MouseButtonFlags, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))
SDL_FUNC("SDL_GetRelativeMouseState", SDL_MouseButtonFlags, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))
SDL_FUNC("SDL_WarpMouseInWindow", None, ctypes.POINTER(SDL_Window), ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_WarpMouseGlobal", ctypes.c_bool, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_SetWindowRelativeMouseMode", ctypes.c_bool, ctypes.POINTER(SDL_Window), ctypes.c_bool)
SDL_FUNC("SDL_GetWindowRelativeMouseMode", ctypes.c_bool, ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_CaptureMouse", ctypes.c_bool, ctypes.c_bool)

SDL_FUNC("SDL_CreateCursor", ctypes.POINTER(SDL_Cursor), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_CreateColorCursor", ctypes.POINTER(SDL_Cursor), ctypes.POINTER(SDL_Surface), ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_CreateSystemCursor", ctypes.POINTER(SDL_Cursor), SDL_SystemCursor)
SDL_FUNC("SDL_SetCursor", ctypes.c_bool, ctypes.POINTER(SDL_Cursor))
SDL_FUNC("SDL_GetCursor", ctypes.POINTER(SDL_Cursor))
SDL_FUNC("SDL_GetDefaultCursor", ctypes.POINTER(SDL_Cursor))
SDL_FUNC("SDL_DestroyCursor", None, ctypes.POINTER(SDL_Cursor))
SDL_FUNC("SDL_ShowCursor", ctypes.c_bool)
SDL_FUNC("SDL_HideCursor", ctypes.c_bool)
SDL_FUNC("SDL_CursorVisible", ctypes.c_bool)


from .__init__ import os, inspect, ctypes, re, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_GET_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_ASSERT_LEVEL = 2
SDL_NULL_WHILE_LOOP_CONDITION = 0

SDL_AssertState = ctypes.c_int

SDL_ASSERTION_RETRY = 0
SDL_ASSERTION_BREAK = 1
SDL_ASSERTION_ABORT = 2
SDL_ASSERTION_IGNORE = 3
SDL_ASSERTION_ALWAYS_IGNORE = 4

SDL_AssertData = ctypes.c_void_p

class SDL_AssertData(ctypes.Structure):
    _fields_ = [
        ("always_ignore", ctypes.c_bool),
        ("trigger_count", ctypes.c_uint),
        ("condition", ctypes.c_char_p),
        ("filename", ctypes.c_char_p),
        ("linenum", ctypes.c_int),
        ("function", ctypes.c_char_p),
        ("next", ctypes.POINTER(SDL_AssertData))
    ]

SDL_FUNC("SDL_ReportAssertion", SDL_AssertState, ctypes.POINTER(SDL_AssertData), ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int)

SDL_TriggerBreakpoint = lambda: breakpoint()
SDL_AssertBreakpoint = lambda: SDL_TriggerBreakpoint()

def SDL_disabled_assert(condition: ctypes.c_bool) -> None:
    """Do not call this function directly."""

def SDL_enabled_assert(condition: ctypes.c_bool) -> None:
    """Do not call this function directly."""

    while not condition:
        data = SDL_AssertData()
        module = current = inspect.currentframe().f_back.f_back
        while module.f_code.co_name != "<module>": module = module.f_back
        match = re.search(r"\((.*?)\)", inspect.getsource(module).split("\n")[current.f_lineno - 1])
        data.condition = (match.group(1) if match else "<condition>").encode()
        state = SDL_GET_BINARY(SDL_BINARY).SDL_ReportAssertion \
            (ctypes.byref(data), current.f_code.co_name.encode(), os.path.split(current.f_code.co_filename)[-1].encode(), current.f_lineno)

        if state in [SDL_ASSERTION_RETRY]:
            continue

        if state in [SDL_ASSERTION_BREAK]:
            return SDL_AssertBreakpoint()

match SDL_ASSERT_LEVEL:
    case 0:
        SDL_assert = lambda condition: SDL_disabled_assert(condition)
        SDL_assert_release = lambda condition: SDL_disabled_assert(condition)
        SDL_assert_paranoid = lambda condition: SDL_disabled_assert(condition)

    case 1:
        SDL_assert = lambda condition: SDL_disabled_assert(condition)
        SDL_assert_release = lambda condition: SDL_enabled_assert(condition)
        SDL_assert_paranoid = lambda condition: SDL_disabled_assert(condition)

    case 2:
        SDL_assert = lambda condition: SDL_enabled_assert(condition)
        SDL_assert_release = lambda condition: SDL_enabled_assert(condition)
        SDL_assert_paranoid = lambda condition: SDL_disabled_assert(condition)

    case 3:
        SDL_assert = lambda condition: SDL_enabled_assert(condition)
        SDL_assert_release = lambda condition: SDL_enabled_assert(condition)
        SDL_assert_paranoid = lambda condition: SDL_enabled_assert(condition)

    case _:
        SDL_enabled_assert(False)

SDL_assert_always = lambda condition: SDL_enabled_assert(condition)
SDL_AssertionHandler = ctypes.CFUNCTYPE(SDL_AssertState, ctypes.POINTER(SDL_AssertData), ctypes.c_void_p)

SDL_FUNC("SDL_SetAssertionHandler", None, SDL_AssertionHandler, ctypes.c_void_p)
SDL_FUNC("SDL_GetDefaultAssertionHandler", SDL_AssertionHandler)
SDL_FUNC("SDL_GetAssertionHandler", SDL_AssertionHandler, ctypes.POINTER(ctypes.c_void_p))
SDL_FUNC("SDL_GetAssertionReport", ctypes.POINTER(SDL_AssertData))
SDL_FUNC("SDL_ResetAssertionReport", None)


from .__init__ import sys, os, ctypes, atexit, typing, types, SDL_GET_BINARY, SDL_BINARY

from .SDL_main import SDL_main_func, \
    SDL_AppEvent_func, SDL_AppInit_func, SDL_AppIterate_func, SDL_AppQuit_func

LP_c_char_p: typing.TypeAlias = ctypes.POINTER(ctypes.c_char_p) # type: ignore

if not int(os.environ.get("SDL_MAIN_HANDLED", "0")) > 0 and not int(os.environ.get("SDL_MAIN_NOIMPL", "0")) > 0:
    import __main__

    if int(os.environ.get("SDL_MAIN_USE_CALLBACKS", "0")) > 0:
        @SDL_main_func
        def SDL_main(argc: ctypes.c_int, argv: LP_c_char_p) -> ctypes.c_int:
            callbacks = [getattr(__main__, i, None) for i in ["SDL_AppInit", "SDL_AppIterate", "SDL_AppEvent", "SDL_AppQuit"]]

            for index, i in enumerate(callbacks):
                if i is not None and not isinstance(i, ctypes._CFuncPtr):
                    callbacks[index] = globals()[f"{i.__name__}_func"](i)

            return SDL_GET_BINARY(SDL_BINARY).SDL_EnterAppMainCallbacks(argc, argv, *callbacks)
        
        os.environ["SDL_MAIN_CALLBACK_STANDARD"] = "1"
        setattr(__main__, "SDL_main", SDL_main)

    if (not int(os.environ.get("SDL_MAIN_USE_CALLBACKS", "0")) > 0 or int(os.environ.get("SDL_MAIN_CALLBACK_STANDARD", "0")) > 0) and not int(os.environ.get("SDL_MAIN_EXPORTED", "0")) > 0:
        @atexit.register
        def SDL_ATEXIT_HANDLER() -> ...:
            if main := (getattr(__main__, "SDL_main", None) or getattr(__main__, "main", None)):
                if not isinstance(main, SDL_main_func):
                    if isinstance(main, types.ModuleType): return
                    if main.__name__ == "main": return
                    main = SDL_main_func(main)

                return SDL_GET_BINARY(SDL_BINARY).SDL_RunApp(len(sys.argv), (ctypes.c_char_p * len(sys.argv))(*[i.encode() for i in sys.argv]), main, None)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_stdinc import SDL_Time

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_FUNC("SDL_GetBasePath", ctypes.c_char_p)
SDL_FUNC("SDL_GetPrefPath", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p)

SDL_Folder = ctypes.c_int

SDL_FOLDER_HOME = 0
SDL_FOLDER_DESKTOP = 1
SDL_FOLDER_DOCUMENTS = 2
SDL_FOLDER_DOWNLOADS = 3
SDL_FOLDER_MUSIC = 4
SDL_FOLDER_PICTURES = 5
SDL_FOLDER_PUBLICSHARE = 6
SDL_FOLDER_SAVEDGAMES = 7
SDL_FOLDER_SCREENSHOTS = 8
SDL_FOLDER_TEMPLATES = 9
SDL_FOLDER_VIDEOS = 10
SDL_FOLDER_COUNT = 11

SDL_FUNC("SDL_GetUserFolder", ctypes.c_char_p, SDL_Folder)

SDL_PathType = ctypes.c_int

SDL_PATHTYPE_NONE = 0
SDL_PATHTYPE_FILE = 1
SDL_PATHTYPE_DIRECTORY = 2
SDL_PATHTYPE_OTHER = 3

class SDL_PathInfo(ctypes.Structure):
    _fields_ = [
        ("type", SDL_PathType),
        ("size", ctypes.c_uint64),
        ("create_time", SDL_Time),
        ("modify_time", SDL_Time),
        ("access_time", SDL_Time)
    ]

SDL_GlobFlags = ctypes.c_uint32

SDL_GLOB_CASEINSENSITIVE = 1 << 0

SDL_FUNC("SDL_CreateDirectory", ctypes.c_bool, ctypes.c_char_p)

SDL_EnumerationResult = ctypes.c_int

SDL_ENUM_CONTINUE = 0
SDL_ENUM_SUCCESS = 1
SDL_ENUM_FAILURE = 2

SDL_EnumerateDirectoryCallback = ctypes.CFUNCTYPE(SDL_EnumerationResult, ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p)

SDL_FUNC("SDL_EnumerateDirectory", ctypes.c_bool, ctypes.c_char_p, SDL_EnumerateDirectoryCallback, ctypes.c_void_p)
SDL_FUNC("SDL_RemovePath", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("SDL_RenamePath", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_CopyFile", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_GetPathInfo", ctypes.c_bool, ctypes.c_char_p, ctypes.POINTER(SDL_PathInfo))
SDL_FUNC("SDL_GlobDirectory", ctypes.POINTER(ctypes.c_char_p), ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(SDL_GlobFlags), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetCurrentDirectory", ctypes.c_char_p)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_video import SDL_Window
from .SDL_stdinc import SDL_FunctionPointer

SDL_SET_CURRENT_BINARY(SDL_BINARY)

class VkInstance(ctypes.c_void_p):
    ...

class VkPhysicalDevice(ctypes.c_void_p):
    ...

class VkAllocationCallbacks(ctypes.c_void_p):
    ...
    
VkSurfaceKHR = ctypes.c_uint64

SDL_FUNC("SDL_Vulkan_LoadLibrary", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("SDL_Vulkan_GetVkGetInstanceProcAddr", SDL_FunctionPointer)
SDL_FUNC("SDL_Vulkan_UnloadLibrary", None)
SDL_FUNC("SDL_Vulkan_GetInstanceExtensions", ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(ctypes.c_uint32))

SDL_FUNC("SDL_Vulkan_CreateSurface", ctypes.c_bool, ctypes.POINTER(SDL_Window), VkInstance, ctypes.POINTER(VkAllocationCallbacks), ctypes.POINTER(VkSurfaceKHR))
SDL_FUNC("SDL_Vulkan_DestroySurface", None, VkInstance, VkSurfaceKHR, ctypes.POINTER(VkAllocationCallbacks))
SDL_FUNC("SDL_Vulkan_GetPresentationSupport", ctypes.c_bool, VkInstance, VkPhysicalDevice, ctypes.c_uint32)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_CACHELINE_SIZE = 128

SDL_FUNC("SDL_GetNumLogicalCPUCores", ctypes.c_int)
SDL_FUNC("SDL_GetCPUCacheLineSize", ctypes.c_int)

SDL_FUNC("SDL_HasAltiVec", ctypes.c_bool)
SDL_FUNC("SDL_HasMMX", ctypes.c_bool)
SDL_FUNC("SDL_HasSSE", ctypes.c_bool)
SDL_FUNC("SDL_HasSSE2", ctypes.c_bool)
SDL_FUNC("SDL_HasSSE3", ctypes.c_bool)
SDL_FUNC("SDL_HasSSE41", ctypes.c_bool)
SDL_FUNC("SDL_HasSSE42", ctypes.c_bool)
SDL_FUNC("SDL_HasAVX", ctypes.c_bool)
SDL_FUNC("SDL_HasAVX2", ctypes.c_bool)
SDL_FUNC("SDL_HasAVX512F", ctypes.c_bool)
SDL_FUNC("SDL_HasARMSIMD", ctypes.c_bool)
SDL_FUNC("SDL_HasNEON", ctypes.c_bool)
SDL_FUNC("SDL_HasLSX", ctypes.c_bool)
SDL_FUNC("SDL_HasLASX", ctypes.c_bool)

SDL_FUNC("SDL_GetSystemRAM", ctypes.c_int)
SDL_FUNC("SDL_GetSIMDAlignment", ctypes.c_size_t)